"""State module for managing Authorization Role Definitions."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    scope: str,
    role_definition_id: str,
    role_definition_name: str,
    permissions: List[
        make_dataclass(
            "ActionsSet",
            [
                ("actions", list, field(default=None)),
                ("notActions", list, field(default=None)),
                ("dataActions", list, field(default=None)),
                ("notDataActions", list, field(default=None)),
            ],
        )
    ],
    description: str = None,
    assignable_scopes: List[str] = None,
    resource_id: str = None,
) -> Dict:
    r"""Create or update Role Definitions.

    Args:
        name(str): The identifier for this state.
        scope(str): The scope of the role definition.
        role_definition_id(str): The ID of the role definition.
        role_definition_name(str): The name of the role definition.
        permissions(list): The permissions of the role definitions.

            * actions(list, Optional):
                Allowed actions
            * notActions(list, Optional):
                Denied actions
            * dataActions(list, Optional):
                Allowed Data actions
            * notDataActions(list, Optional):
                Denied Data actions
        description(str): The description of the role definitions.
        assignable_scopes(list, Optional): The assignable scopes of the role definitions. Defaults to scope.
        resource_id(str, Optional): Role definition resource id on Azure.

    Returns:
        dict

    Examples:
        .. code-block:: sls

            my-role-definition:
              azure.authorization.role_definitions.present:
                - name: my-role-definition
                - scope: /subscriptions/subscription-id
                - role_definition_id: aa246b4f-3ba8-4d39-8e43-687eb2e1661c
                - role_definition_name: my-role-name
                - permissions:
                  - actions:
                    - Microsoft.Resources/subscriptions/resourceGroups/read
                  - notActions:
                    - Microsoft.Resources/subscriptions/resourceGroups/write
                  - dataActions:
                    - Microsoft.Storage/storageAccounts/blobServices/containers/blobs/read
                  - notDataActions:
                    - Microsoft.Storage/storageAccounts/blobServices/containers/blobs/write
                - description: my-role-description
                - assignable_scopes:
                  - /subscriptions/subscription-id
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if resource_id is None:
        resource_id = f"{scope}/providers/Microsoft.Authorization/roleDefinitions/{role_definition_id}"
    response_get = await hub.exec.azure.authorization.role_definitions.get(
        ctx, resource_id=resource_id, raw=True
    )

    if not response_get["result"]:
        hub.log.debug(
            f"Could not get azure.authorization.role_definitions {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"].extend(
            hub.tool.azure.result_utils.extract_error_comments(response_get)
        )
        return result

    if not response_get["ret"]:
        if ctx.get("test", False):
            # Return a proposed state by Idem state --test
            result["new_state"] = hub.tool.azure.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "scope": scope,
                    "role_definition_id": role_definition_id,
                    "resource_id": resource_id,
                    "role_definition_name": role_definition_name,
                    "description": description,
                    "permissions": permissions,
                    "assignable_scopes": assignable_scopes,
                },
            )
            result["comment"].append(
                f"Would create azure.authorization.role_definitions '{name}'"
            )
            return result
        else:
            # PUT operation to create a resource
            payload = hub.tool.azure.authorization.role_definitions.convert_present_to_raw_role_definitions(
                scope=scope,
                role_definition_name=role_definition_name,
                description=description,
                permissions=permissions,
                assignable_scopes=assignable_scopes,
            )
            response_put = await hub.exec.request.json.put(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2022-04-01",
                success_codes=[201],
                json=payload,
            )

            if not response_put["result"]:
                hub.log.debug(
                    f"Could not create azure.authorization.role_definitions {response_put['comment']} {response_put['ret']}"
                )
                result["comment"].extend(
                    hub.tool.azure.result_utils.extract_error_comments(response_put)
                )
                result["result"] = False
                return result
            result[
                "new_state"
            ] = hub.tool.azure.authorization.role_definitions.convert_raw_role_definitions_to_present(
                resource=response_put["ret"],
                idem_resource_name=name,
                role_definition_id=role_definition_id,
                scope=scope,
                resource_id=resource_id,
            )
            result["comment"].append(
                f"Created azure.authorization.role_definitions '{name}'"
            )
            return result
    else:
        existing_resource = response_get["ret"]
        result[
            "old_state"
        ] = hub.tool.azure.authorization.role_definitions.convert_raw_role_definitions_to_present(
            resource=existing_resource,
            idem_resource_name=name,
            role_definition_id=role_definition_id,
            scope=scope,
            resource_id=resource_id,
        )
        new_payload = hub.tool.azure.authorization.role_definitions.update_role_definitions_payload(
            existing_resource,
            {
                "role_definition_name": role_definition_name,
                "description": description,
                "permissions": permissions,
                "assignable_scopes": assignable_scopes,
            },
        )
        if ctx.get("test", False):
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.authorization.role_definitions '{name}' has no property need to be updated."
                )
            else:
                result[
                    "new_state"
                ] = hub.tool.azure.authorization.role_definitions.convert_raw_role_definitions_to_present(
                    resource=new_payload["ret"],
                    idem_resource_name=name,
                    role_definition_id=role_definition_id,
                    scope=scope,
                    resource_id=resource_id,
                )
                result["comment"].append(
                    f"Would update azure.authorization.role_definitions '{name}'"
                )
            return result

        if new_payload["ret"] is None:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                f"azure.authorization.role_definitions '{name}' has no property need to be updated."
            )
            return result
        result["comment"].extend(new_payload["comment"])
        response_put = await hub.exec.request.json.put(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2022-04-01",
            success_codes=[200, 201],
            json=new_payload["ret"],
        )
        if not response_put["result"]:
            hub.log.debug(
                f"Could not update azure.authorization.role_definitions {response_put['comment']} {response_put['ret']}"
            )
            result["result"] = False
            result["comment"].extend(
                hub.tool.azure.result_utils.extract_error_comments(response_put)
            )
            return result

        result[
            "new_state"
        ] = hub.tool.azure.authorization.role_definitions.convert_raw_role_definitions_to_present(
            resource=response_put["ret"],
            idem_resource_name=name,
            role_definition_id=role_definition_id,
            scope=scope,
            resource_id=resource_id,
        )
        result["comment"].append(
            f"Updated azure.authorization.role_definitions '{name}'"
        )
        return result


async def absent(hub, ctx, name: str, scope: str, role_definition_id: str) -> Dict:
    r"""Delete Role Definitions.

    Args:
        name(str): The identifier for this state.
        scope(str): The scope of the role definition.
        role_definition_id(str): The ID of the role definition to delete.

    Returns:
        dict

    Examples:
        .. code-block:: sls

            my-role-definition:
              azure.authorization.role_definitions.absent:
                - name: my-role-definition
                - scope: /subscriptions/subscription-id
                - role_definition_id: aa246b4f-3ba8-4d39-8e43-687eb2e1661c
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    resource_id = f"{scope}/providers/Microsoft.Authorization/roleDefinitions/{role_definition_id}"
    response_get = await hub.exec.azure.authorization.role_definitions.get(
        ctx,
        resource_id=resource_id,
    )
    if not response_get["result"]:
        hub.log.debug(
            f"Could not get azure.authorization.role_definitions {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"].extend(
            hub.tool.azure.result_utils.extract_error_comments(response_get)
        )
        return result

    if response_get["ret"]:
        result["old_state"] = response_get["ret"]
        result["old_state"]["name"] = name
        if ctx.get("test", False):
            result["comment"].append(
                f"Would delete azure.authorization.role_definitions '{name}'"
            )
            return result

        response_delete = await hub.exec.request.raw.delete(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2022-04-01",
            success_codes=[200, 204],
        )
        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete azure.authorization.role_definitions {response_delete['comment']} {response_delete['ret']}"
            )
            result["result"] = False
            result["comment"].extend(
                hub.tool.azure.result_utils.extract_error_comments(response_delete)
            )
            return result

        result["comment"].append(
            f"Deleted azure.authorization.role_definitions '{name}'"
        )
        return result
    else:
        # If Azure returns 'Not Found' error, it means the resource has been absent.
        result["comment"].append(
            f"azure.authorization.role_definitions '{name}' already absent"
        )
        return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all Role Definitions under the same subscription.


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.authorization.role_definitions
    """
    result = {}
    ret_list = await hub.exec.azure.authorization.role_definitions.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(
            f"Could not describe authorization role_definitions {ret_list['comment']}"
        )
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.authorization.role_definitions.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



from typing import Any
from typing import Dict


def update_result_for_unsuccessful_operation(
    hub,
    result: Dict[str, Any],
    resource_type: str,
    name: str,
    response: Dict[str, Any],
    operation: str,
):
    result["result"] = False
    result["comment"].append(
        hub.tool.azure.comment_utils.could_not_do_operation_comment(
            resource_type, name, operation
        )
    )

    if response.get("comment"):
        comment = response["comment"]
        if isinstance(comment, str):
            result["comment"].append(comment)
        elif isinstance(comment, list):
            result["comment"].extend(comment)

    if response.get("ret"):
        result["comment"].append(str(response["ret"]))

    hub.log.debug(f"{result['comment']}")
    return result


def absent_implemented_through_wrapper_result(hub, resource_type: str, name: str):
    return {
        "comment": [
            f"Invalid invocation! Absent method for {resource_type} is implemented by wrapper."
        ],
        "old_state": None,
        "new_state": None,
        "name": name,
        "result": False,
    }


def extract_error_comments(hub, response: Dict[str, Any]):
    comment = []
    if response.get("comment"):
        comment.append(response["comment"])
    if isinstance(response.get("ret"), str):
        comment.append(response["ret"])

    return comment



"""State module for managing SQL Databases."""

__contracts__ = ["resource"]

import copy
import uuid
from dataclasses import make_dataclass, field
from typing import Dict, Any


async def present(
    hub,
    ctx,
    name: str,
    location: str,
    resource_id: str = None,
    subscription_id: str = None,
    resource_group_name: str = None,
    server_name: str = None,
    database_name: str = None,
    identity: make_dataclass(
        "DatabaseIdentity",
        [
            ("tenant_id", str, field(default=None)),
            ("type", str, field(default=None)),
            (
                "user_assigned_identities",
                Dict[
                    str,
                    make_dataclass(
                        "DatabaseUserIdentity",
                        [
                            ("principal_id", str, field(default=None)),
                            ("client_id", str, field(default=None)),
                        ],
                    ),
                ],
                field(default=None),
            ),
        ],
    ) = None,
    auto_pause_delay: int = None,
    catalog_collation: str = None,
    collation: str = None,
    create_mode: str = None,
    elastic_pool_id: str = None,
    federated_client_id: str = None,
    high_availability_replica_count: int = None,
    is_ledger_on: bool = None,
    license_type: str = None,
    long_term_retention_backup_resource_id: str = None,
    maintenance_configuration_id: str = None,
    max_size_bytes: int = None,
    min_capacity: float = None,
    read_scale: str = None,
    recoverable_database_id: str = None,
    recovery_services_recovery_point_id: str = None,
    requested_backup_storage_redundancy: str = None,
    restorable_dropped_database_id: str = None,
    restore_point_in_time: str = None,
    sample_name: str = None,
    secondary_type: str = None,
    source_database_deletion_date: str = None,
    source_database_id: str = None,
    source_resource_id: str = None,
    zone_redundant: bool = None,
    sku: make_dataclass(
        "Sku",
        [
            ("capacity", int, field(default=None)),
            ("family", str, field(default=None)),
            ("name", str, field(default=None)),
            ("size", str, field(default=None)),
            ("tier", str, field(default=None)),
        ],
    ) = None,
    tags: Dict = None,
) -> Dict:
    r"""Create or update SQL Databases.

    Args:
        name(str):
            The identifier for this state.

        resource_id(str, Optional):
            SQL Database resource id on Azure

        resource_group_name(str, Optional):
            The name of the resource group that contains the resource. You can obtain this value from the Azure Resource Manager API or the portal.

        server_name(str, Optional):
            The name of the server.

        database_name(str, Optional):
            The name of the database.

        subscription_id(str, Optional):
            The subscription ID that identifies an Azure subscription.

        location(str):
            Resource location.

        identity(Dict[str, Any], Optional):
            The Azure Active Directory identity of the database.

            * tenant_id(str, Optional):
                The Azure Active Directory tenant id.
            * type(str, Optional):
                The identity type
                    "None"
                    "UserAssigned"
            * user_assigned_identities(Dict[str, DatabaseUserIdentity], Optional):
                The resource ids of the user assigned identities to use. The values are Azure Active Directory identity configuration for a resource:
                    * client_id(str):
                        The Azure Active Directory client id.
                    * principal_id(str):
                        The Azure Active Directory principal id.

        auto_pause_delay(int, Optional):
            Time in minutes after which database is automatically paused. A value of -1 means that automatic pause is disabled

        catalog_collation(str, Optional):
            Collation of the metadata catalog.

        collation(str, Optional):
            The collation of the database.

        create_mode(str, Optional):
            Specifies the mode of database creation.
                "Default": regular database creation.
                "Copy": creates a database as a copy of an existing database. sourceDatabaseId must be specified as the resource ID of the source database.
                "Secondary": creates a database as a secondary replica of an existing database. sourceDatabaseId must be specified as the resource ID of the existing primary database.
                "PointInTimeRestore": Creates a database by restoring a point in time backup of an existing database. sourceDatabaseId must be specified as the resource ID of the existing database, and restorePointInTime must be specified.
                "Recovery": Creates a database by restoring a geo-replicated backup. sourceDatabaseId must be specified as the recoverable database resource ID to restore.
                "Restore": Creates a database by restoring a backup of a deleted database. sourceDatabaseId must be specified. If sourceDatabaseId is the database's original resource ID, then sourceDatabaseDeletionDate must be specified. Otherwise sourceDatabaseId must be the restorable dropped database resource ID and sourceDatabaseDeletionDate is ignored. restorePointInTime may also be specified to restore from an earlier point in time.
                "RestoreLongTermRetentionBackup": Creates a database by restoring from a long term retention vault. recoveryServicesRecoveryPointResourceId must be specified as the recovery point resource ID.

                Copy, Secondary, and RestoreLongTermRetentionBackup are not supported for DataWarehouse edition.

        elastic_pool_id(str, Optional):
            The resource identifier of the elastic pool containing this database.

        federated_client_id(str, Optional):
            The Client id used for cross tenant per database CMK scenario

        high_availability_replica_count(int, Optional):
            The number of secondary replicas associated with the database that are used to provide high availability. Not applicable to a Hyperscale database within an elastic pool.

        is_ledger_on(bool, Optional):
            Whether or not this database is a ledger database, which means all tables in the database are ledger tables. Note: the value of this property cannot be changed after the database has been created.

        license_type(str, Optional):
            The license type to apply for this database. LicenseIncluded if you need a license, or BasePrice if you have a license and are eligible for the Azure Hybrid Benefit.

        long_term_retention_backup_resource_id(str, Optional):
            The resource identifier of the long term retention backup associated with create operation of this database.

        maintenance_configuration_id(str, Optional):
            Maintenance configuration id assigned to the database. This configuration defines the period when the maintenance updates will occur.

        max_size_bytes(int, Optional):
            The max size of the database expressed in bytes.

        min_capacity(float, Optional):
            Minimal capacity that database will always have allocated, if not paused

        read_scale(str, Optional):
            The state of read-only routing. If enabled, connections that have application intent set to readonly in their connection string may be routed to a readonly secondary replica in the same region. Not applicable to a Hyperscale database within an elastic pool.

        recoverable_database_id(str, Optional):
            The resource identifier of the recoverable database associated with create operation of this database.

        recovery_services_recovery_point_id(str, Optional):
            The resource identifier of the recovery point associated with create operation of this database.

        requested_backup_storage_redundancy(str, Optional):
            The storage account type to be used to store backups for this database.

        restorable_dropped_database_id(str, Optional):
            The resource identifier of the restorable dropped database associated with create operation of this database.

        restore_point_in_time(str, Optional):
            Specifies the point in time (ISO8601 format) of the source database that will be restored to create the new database.

        sample_name(str, Optional):
            The name of the sample schema to apply when creating this database.

        secondary_type(str, Optional):
            The secondary type of the database if it is a secondary. Valid values are Geo and Named.

        source_database_deletion_date(str, Optional):
            Specifies the time that the database was deleted.

        source_database_id(str, Optional):
            The resource identifier of the source database associated with create operation of this database.

        source_resource_id(str, Optional):
            The resource identifier of the source associated with the create operation of this database.
            This property is only supported for DataWarehouse edition and allows to restore across subscriptions.
            When sourceResourceId is specified, sourceDatabaseId, recoverableDatabaseId, restorableDroppedDatabaseId and sourceDatabaseDeletionDate must not be specified and CreateMode must be PointInTimeRestore, Restore or Recover.
            When createMode is PointInTimeRestore, sourceResourceId must be the resource ID of the existing database or existing sql pool, and restorePointInTime must be specified.
            When createMode is Restore, sourceResourceId must be the resource ID of restorable dropped database or restorable dropped sql pool.
            When createMode is Recover, sourceResourceId must be the resource ID of recoverable database or recoverable sql pool.
            When source subscription belongs to a different tenant than target subscription, "x-ms-authorization-auxiliary" header must contain authentication token for the source tenant. For more details about "x-ms-authorization-auxiliary" header see https://docs.microsoft.com/en-us/azure/azure-resource-manager/management/authenticate-multi-tenant

        zone_redundant(bool, Optional):
            Whether or not this database is zone redundant, which means the replicas of this database will be spread across multiple availability zones.

        sku(Dict[str, Any], Optional):
            The database SKU.

            * capacity(int, Optional):
                Capacity of the particular SKU.
            * family(str, Optional):
                If the service has different generations of hardware, for the same SKU, then that can be captured here.
            * name(str, Optional):
                The name of the SKU, typically, a letter + Number code, e.g. P3.
            * size(str, Optional):
                Size of the particular SKU
            * tier(str, Optional):
                The tier or edition of the particular SKU, e.g. Basic, Premium.

            The list of SKUs may vary by region and support offer. To determine the SKUs (including the SKU name, tier/edition, family, and capacity) that are available to your subscription in an Azure region, use the Capabilities_ListByLocation REST API or one of the following commands:
                Azure CLI
                az sql db list-editions -l <location> -o table

                PowerShell
                Get-AzSqlServerServiceObjective -Location <location>

        tags(Dict, Optional):
            Resource tags.

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            my-sqldb:
              azure.sql_databases.database.present:
                - location: westeurope
                - subscription_id: 11111111-2222-3333-4444-555555555555
                - resource_group_name: default-rg
                - server_name: my-server-1
                - database_name: test-sqldb-1
                - sku:
                    name: Standard
                    tier: Standard
                    capacity: 10
                - collation: SQL_Latin1_General_CP1_CI_AS
                - max_size_bytes: 268435456000
                - read_scale: Disabled
                - requested_backup_storage_redundancy: Local
    """
    result = ctx.get("wrapper_result")

    if ctx.get("skip_present"):
        return result

    if not result:
        error_message = hub.tool.azure.comment_utils.no_result_from_wrapper(
            "azure.sql_database.databases", name
        )
        hub.log.error(error_message)
        return {
            "result": False,
            "old_state": None,
            "new_state": None,
            "name": name,
            "comment": [error_message],
        }

    computed = ctx.get("computed")
    computed_resource_id = computed.get("resource_id")
    computed_resource_url = computed.get("resource_url")
    existing_resource = result["old_state"]

    if not existing_resource:
        # Create
        if ctx.get("test", False):
            # Return a proposed state by Idem state --test
            result["new_state"] = hub.tool.azure.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "location": location,
                    "resource_id": computed_resource_id,
                    "subscription_id": subscription_id,
                    "resource_group_name": resource_group_name,
                    "server_name": server_name,
                    "database_name": database_name,
                    "identity": identity,
                    "auto_pause_delay": auto_pause_delay,
                    "catalog_collation": catalog_collation,
                    "collation": collation,
                    "create_mode": create_mode,
                    "elastic_pool_id": elastic_pool_id,
                    "federated_client_id": federated_client_id,
                    "high_availability_replica_count": high_availability_replica_count,
                    "is_ledger_on": is_ledger_on,
                    "license_type": license_type,
                    "long_term_retention_backup_resource_id": long_term_retention_backup_resource_id,
                    "maintenance_configuration_id": maintenance_configuration_id,
                    "max_size_bytes": max_size_bytes,
                    "min_capacity": min_capacity,
                    "read_scale": read_scale,
                    "recoverable_database_id": recoverable_database_id,
                    "recovery_services_recovery_point_id": recovery_services_recovery_point_id,
                    "requested_backup_storage_redundancy": requested_backup_storage_redundancy,
                    "restorable_dropped_database_id": restorable_dropped_database_id,
                    "restore_point_in_time": restore_point_in_time,
                    "sample_name": sample_name,
                    "secondary_type": secondary_type,
                    "source_database_deletion_date": source_database_deletion_date,
                    "source_database_id": source_database_id,
                    "source_resource_id": source_resource_id,
                    "zone_redundant": zone_redundant,
                    "sku": sku,
                    "tags": tags,
                },
            )
            result["comment"].append(
                f"Would create azure.sql_database.databases '{name}'"
            )
            return result
        else:
            # PUT operation to create a resource
            payload = hub.tool.azure.sql_database.databases.convert_present_to_raw_database(
                subscription_id=subscription_id,
                name=name,
                location=location,
                identity=identity,
                auto_pause_delay=auto_pause_delay,
                catalog_collation=catalog_collation,
                collation=collation,
                create_mode=create_mode,
                elastic_pool_id=elastic_pool_id,
                federated_client_id=federated_client_id,
                high_availability_replica_count=high_availability_replica_count,
                is_ledger_on=is_ledger_on,
                license_type=license_type,
                long_term_retention_backup_resource_id=long_term_retention_backup_resource_id,
                maintenance_configuration_id=maintenance_configuration_id,
                max_size_bytes=max_size_bytes,
                min_capacity=min_capacity,
                read_scale=read_scale,
                recoverable_database_id=recoverable_database_id,
                recovery_services_recovery_point_id=recovery_services_recovery_point_id,
                requested_backup_storage_redundancy=requested_backup_storage_redundancy,
                restorable_dropped_database_id=restorable_dropped_database_id,
                restore_point_in_time=restore_point_in_time,
                sample_name=sample_name,
                secondary_type=secondary_type,
                source_database_deletion_date=source_database_deletion_date,
                source_database_id=source_database_id,
                source_resource_id=source_resource_id,
                zone_redundant=zone_redundant,
                sku=sku,
                tags=tags,
            )
            response_put = await hub.exec.request.json.put(
                ctx,
                url=f"{ctx.computed.resource_url}",
                success_codes=[200, 201, 202],
                json=payload,
            )

            if response_put["status"] == 202:
                # Creating the database is in progress.
                # TODO: Use reconciler mechanism to wait for operation to complete
                result["rerun_data"] = {
                    "operation_id": str(uuid.uuid4()),
                    "operation_headers": dict(response_put.get("headers")),
                    "resource_id": f"{computed_resource_id}",
                    "resource_url": f"{computed_resource_url}",
                    "old_state": result["old_state"],
                }
                return result
            elif not response_put["result"] and response_put["status"] != 202:
                hub.log.debug(
                    f"Could not create azure.sql_database.databases {response_put['comment']} {response_put['ret']}"
                )
                result["comment"].extend(
                    hub.tool.azure.result_utils.extract_error_comments(response_put)
                )
                result["result"] = False
                return result
            else:
                result[
                    "new_state"
                ] = hub.tool.azure.sql_database.databases.convert_raw_database_to_present(
                    resource=response_put["ret"],
                    idem_resource_name=name,
                    resource_id=computed_resource_id,
                )

            result["comment"].append(f"Created azure.sql_database.databases '{name}'")
            return result

    else:
        # Update
        # Generate a new PUT operation payload with new values
        new_payload = hub.tool.azure.sql_database.databases.update_database_payload(
            existing_resource,
            {
                "id": resource_id,
                "location": location,
                "identity": identity,
                "auto_pause_delay": auto_pause_delay,
                "catalog_collation": catalog_collation,
                "collation": collation,
                "create_mode": create_mode,
                "elastic_pool_id": elastic_pool_id,
                "federated_client_id": federated_client_id,
                "high_availability_replica_count": high_availability_replica_count,
                "is_ledger_on": is_ledger_on,
                "license_type": license_type,
                "long_term_retention_backup_resource_id": long_term_retention_backup_resource_id,
                "maintenance_configuration_id": maintenance_configuration_id,
                "max_size_bytes": max_size_bytes,
                "min_capacity": min_capacity,
                "read_scale": read_scale,
                "recoverable_database_id": recoverable_database_id,
                "recovery_services_recovery_point_id": recovery_services_recovery_point_id,
                "requested_backup_storage_redundancy": requested_backup_storage_redundancy,
                "restorable_dropped_database_id": restorable_dropped_database_id,
                "restore_point_in_time": restore_point_in_time,
                "sample_name": sample_name,
                "secondary_type": secondary_type,
                "source_database_deletion_date": source_database_deletion_date,
                "source_database_id": source_database_id,
                "source_resource_id": source_resource_id,
                "zone_redundant": zone_redundant,
                "sku": sku,
                "tags": tags,
            },
        )
        if ctx.get("test", False):
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    hub.tool.azure.comment_utils.no_property_to_be_updated_comment(
                        f"azure.sql_database.databases", name
                    )
                )
            else:
                result[
                    "new_state"
                ] = hub.tool.azure.sql_database.databases.convert_raw_database_to_present(
                    resource=new_payload["ret"],
                    idem_resource_name=name,
                    resource_id=computed_resource_id,
                )
                result["comment"].append(
                    f"Would update azure.sql_database.databases '{name}'"
                )
            return result
        # PUT operation to update a resource
        if new_payload["ret"] is None:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                hub.tool.azure.comment_utils.no_property_to_be_updated_comment(
                    f"azure.sql_database.databases", name
                )
            )
            return result
        result["comment"].extend(new_payload["comment"])
        response_put = await hub.exec.request.json.put(
            ctx,
            url=f"{ctx.computed.resource_url}",
            success_codes=[200, 201, 202],
            json=new_payload["ret"],
        )

        if not response_put["result"]:
            hub.log.debug(
                f"Could not update azure.sql_database.databases {response_put['comment']} {response_put['ret']}"
            )
            result["result"] = False
            result["comment"].extend(
                hub.tool.azure.result_utils.extract_error_comments(response_put)
            )
            return result

        if response_put["status"] == 202:
            # Updating the database is in progress.
            # TODO: Use reconciler mechanism to wait for operation to complete
            result["rerun_data"] = {
                "operation_id": str(uuid.uuid4()),
                "operation_headers": dict(response_put.get("headers")),
                "resource_id": f"{computed_resource_id}",
                "resource_url": f"{computed_resource_url}",
                "old_state": result["old_state"],
            }
            return result
        elif not response_put["result"] and response_put["status"] != 202:
            hub.log.debug(
                f"Could not update azure.sql_database.databases {response_put['comment']} {response_put['ret']}"
            )
            result["comment"].extend(
                hub.tool.azure.result_utils.extract_error_comments(response_put)
            )
            result["result"] = False
            return result
        else:
            result[
                "new_state"
            ] = hub.tool.azure.sql_database.databases.convert_raw_database_to_present(
                resource=response_put["ret"],
                idem_resource_name=name,
                resource_id=computed_resource_id,
            )

        result["comment"].append(f"Updated azure.sql_database.databases '{name}'")
        return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_group_name: str = None,
    server_name: str = None,
    database_name: str = None,
    subscription_id: str = None,
    resource_id: str = None,
) -> Dict:
    r"""Delete SQL databases.

    Args:
        name(str):
            The identifier for this state.

        resource_group_name(str, Optional):
            The name of the resource group.

        server_name(str, Optional):
            The name of the server.

        database_name(str, Optional):
            The name of the database.

        subscription_id(str, Optional):
            Subscription Unique id.

        resource_id(str, Optional): An identifier of the resource in the provider. Defaults to None.

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.sql_database.databases.absent:
                - name: value
                - resource_group_name: value
                - server_name: value
                - database_name: value
    """

    # Resource deletion is handled via common recursive_contracts.init#call_absent() wrapper
    raise NotImplementedError


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all databases under the same subscription.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.sql_database.databases
    """
    result = {}
    ret_list = await hub.exec.azure.sql_database.databases.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(f"Could not describe SQL database {ret_list['comment']}")
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.sql_database.databases.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """Default implemented for each module."""
    return hub.tool.azure.resource_utils.is_pending(
        ret=ret, state=state, **pending_kwargs
    )



"""State module for managing Virtual Network."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    virtual_network_name: str,
    address_space: List[str],
    location: str,
    resource_id: str = None,
    extended_location: make_dataclass(
        "ExtendedLocation",
        [
            ("name", str),
            ("type", str),
        ],
    ) = None,
    bgp_communities: make_dataclass(
        "VirtualNetworkBgpCommunities",
        [
            ("virtual_network_community", str),
        ],
    ) = None,
    flow_timeout_in_minutes: int = None,
    subnets: List[
        make_dataclass(
            "SubnetSet",
            [
                ("name", str, field(default=None)),
                ("address_prefix", str, field(default=None)),
                ("address_prefixes", List, field(default=None)),
                ("security_group_id", str, field(default=None)),
                ("service_endpoints", List, field(default=None)),
            ],
        )
    ] = None,
    ddos_protection_plan: make_dataclass(
        "Id",
        [
            ("id", str),
        ],
    ) = None,
    enable_ddos_protection: bool = None,
    enable_vm_protection: bool = None,
    dhcp_options: make_dataclass(
        "DhcpOptions", [("dns_servers", List[str], field(default=None))]
    ) = None,
    subscription_id: str = None,
    tags: Dict = None,
) -> Dict:
    r"""Create or update Virtual Networks.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        virtual_network_name(str): The name of the virtual network.
        address_space(list): An array of IP address ranges that can be used by subnets of the virtual network.
        location(str): Resource location. This field can not be updated.
        resource_id(str, Optional): Virtual Network resource id on Azure.
        extended_location(Dict, Optional): The extended location of the virtual network.

            * name(str):
                The name of the extended location.
            * type(str):
                The type of the extended location.
                Enum type. Allowed values: EdgeZone
        bgp_communities(Dict, Optional): Bgp Communities sent over ExpressRoute with each route corresponding to a prefix in this VNET.

            * virtual_network_community(str):
                The BGP community associated with the virtual network.
        flow_timeout_in_minutes(int, Optional): The FlowTimeout value (in minutes) for the Virtual Network
        subnets(list, Optional): List of Subnet in a virtual network resource.Each Subnet will have fields

            * name(str, Optional):
                The name of subnet.
            * address_prefix(str, Optional):
                The address prefix for the subnet.
            * address_prefixes(list, Optional):
                List of address prefixes for the subnet.
            * security_group_id(str, Optional):
                The security group id.
            * service_endpoints(list, Optional):
                List of service endpoint.
        dhcp_options(Dict, Optional): DhcpOptions contains an array of DNS servers available to VMs deployed in the virtual network. Standard DHCP option for a subnet overrides VNET DHCP options.

            * dns_servers(List, Optional):
                The list of DNS servers IP addresses.
        ddos_protection_plan(Dict, Optional):

            * id(str):
                The DDoS protection plan associated with the virtual network.
        enable_ddos_protection(bool, Optional): Indicates if DDoS protection is enabled for all the protected resources in the virtual network. It requires a DDoS protection plan associated with the resource.
        enable_vm_protection(bool, Optional): Indicates if VM protection is enabled for all the subnets in the virtual network.
        subscription_id(str, Optional): Subscription Unique id.
        tags(dict, Optional): Resource tags.

    Returns:
        dict

    Examples:
        .. code-block:: sls

            my-vnet:
              azure.network.virtual_networks.present:
                - name: my-vnet
                - resource_group_name: my-rg-1
                - virtual_network_name: my-vnet-1
                - location: westus
                - flow_timeout_in_minutes: 15
                - tags:
                    my-tag-key: my-tag-value
                - subnets:
                    - name: subnet_name
                      address_prefix: 10.0.0.0/26
                      security_group_id: /subscriptions/subscription_id/resourceGroups/resource_group_name/providers/Microsoft.Network/networkSecurityGroups/network-security-group-name
                      service_endpoints:
                          - Microsoft.Storage
                - address_space:
                    - 10.0.0.0/26
    """
    result = ctx.get("wrapper_result")

    if ctx.get("skip_present"):
        return result

    if not result:
        error_message = hub.tool.azure.comment_utils.no_result_from_wrapper(
            "azure.network.virtual_networks", name
        )
        hub.log.error(error_message)
        return {
            "result": False,
            "old_state": None,
            "new_state": None,
            "name": name,
            "comment": [error_message],
        }

    computed = ctx.get("computed")
    computed_resource_id = computed.get("resource_id")
    computed_resource_url = computed.get("resource_url")
    existing_resource = result["old_state"]

    if not existing_resource:
        if ctx.get("test", False):
            # Return a proposed state by Idem state --test
            result["new_state"] = hub.tool.azure.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_group_name": resource_group_name,
                    "virtual_network_name": virtual_network_name,
                    "address_space": address_space,
                    "tags": tags,
                    "location": location,
                    "resource_id": computed_resource_id,
                    "extended_location": extended_location,
                    "flow_timeout_in_minutes": flow_timeout_in_minutes,
                    "bgp_communities": bgp_communities,
                    "subnets": subnets,
                    "ddos_protection_plan": ddos_protection_plan,
                    "enable_ddos_protection": enable_ddos_protection,
                    "enable_vm_protection": enable_vm_protection,
                    "dhcp_options": dhcp_options,
                    "subscription_id": subscription_id,
                },
            )
            result["comment"].append(
                f"Would create azure.network.virtual_networks '{name}'"
            )
            return result
        else:
            # PUT operation to create a resource
            payload = hub.tool.azure.network.virtual_networks.convert_present_to_raw_virtual_network(
                subscription_id=subscription_id,
                address_space=address_space,
                location=location,
                extended_location=extended_location,
                bgp_communities=bgp_communities,
                flow_timeout_in_minutes=flow_timeout_in_minutes,
                subnets=subnets,
                ddos_protection_plan=ddos_protection_plan,
                enable_ddos_protection=enable_ddos_protection,
                enable_vm_protection=enable_vm_protection,
                dhcp_options=dhcp_options,
                tags=tags,
            )
            response_put = await hub.exec.request.json.put(
                ctx,
                url=computed_resource_url,
                success_codes=[200, 201],
                json=payload,
            )

            if not response_put["result"]:
                hub.log.debug(
                    f"Could not create azure.network.virtual_networks {response_put['comment']} {response_put['ret']}"
                )
                result["comment"].extend(
                    hub.tool.azure.result_utils.extract_error_comments(response_put)
                )
                result["result"] = False
                return result

            result[
                "new_state"
            ] = hub.tool.azure.network.virtual_networks.convert_raw_virtual_network_to_present(
                resource=response_put["ret"],
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                virtual_network_name=virtual_network_name,
                resource_id=computed_resource_id,
                subscription_id=subscription_id,
            )
            result["comment"].append(f"Created azure.network.virtual_networks '{name}'")
            return result

    else:
        raw_existing_resource = hub.tool.azure.network.virtual_networks.convert_present_to_raw_virtual_network(
            **existing_resource
        )
        # Generate a new PUT operation payload with new values
        new_payload = (
            hub.tool.azure.network.virtual_networks.update_virtual_network_payload(
                subscription_id,
                raw_existing_resource,
                {
                    "extended_location": extended_location,
                    "address_space": address_space,
                    "bgp_communities": bgp_communities,
                    "flow_timeout_in_minutes": flow_timeout_in_minutes,
                    "subnets": subnets,
                    "ddos_protection_plan": ddos_protection_plan,
                    "enable_ddos_protection": enable_ddos_protection,
                    "enable_vm_protection": enable_vm_protection,
                    "dhcp_options": dhcp_options,
                    "tags": tags,
                },
            )
        )
        if ctx.get("test", False):
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    hub.tool.azure.comment_utils.no_property_to_be_updated_comment(
                        "azure.network.virtual_networks", name
                    )
                )
            else:
                result[
                    "new_state"
                ] = hub.tool.azure.network.virtual_networks.convert_raw_virtual_network_to_present(
                    resource=new_payload["ret"],
                    idem_resource_name=name,
                    resource_group_name=resource_group_name,
                    virtual_network_name=virtual_network_name,
                    resource_id=computed_resource_id,
                    subscription_id=subscription_id,
                )
                result["comment"].append(
                    f"Would update azure.network.virtual_networks '{name}'"
                )
            return result
        # PUT operation to update a resource
        if new_payload["ret"] is None:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                hub.tool.azure.comment_utils.no_property_to_be_updated_comment(
                    "azure.network.virtual_networks", name
                )
            )
            return result
        result["comment"].extend(new_payload["comment"])
        response_put = await hub.exec.request.json.put(
            ctx,
            url=computed_resource_url,
            success_codes=[200],
            json=new_payload["ret"],
        )

        if not response_put["result"]:
            hub.log.debug(
                f"Could not update azure.network.virtual_networks {response_put['comment']} {response_put['ret']}"
            )
            result["result"] = False
            result["comment"].extend(
                hub.tool.azure.result_utils.extract_error_comments(response_put)
            )
            return result

        result[
            "new_state"
        ] = hub.tool.azure.network.virtual_networks.convert_raw_virtual_network_to_present(
            resource=response_put["ret"],
            idem_resource_name=name,
            resource_group_name=resource_group_name,
            virtual_network_name=virtual_network_name,
            resource_id=computed_resource_id,
            subscription_id=subscription_id,
        )
        result["comment"].append(f"Updated azure.network.virtual_networks '{name}'")
        return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    resource_group_name: str = None,
    virtual_network_name: str = None,
    subscription_id: str = None,
) -> Dict:
    r"""Delete Virtual Networks.

    Args:
        name(str): The identifier for this state.
        resource_id(str, Optional): Virtual Network resource id on Azure.
        resource_group_name(str, Optional): The name of the resource group.
        virtual_network_name(str, Optional): The name of the virtual network.
        subscription_id(str, Optional): Subscription Unique id.

    Returns:
        dict

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.network.virtual_networks.absent:
                - name: value
                - resource_group_name: value
                - virtual_network_name: value
    """
    # Resource deletion is handled via common recursive_contracts.init#call_absent() wrapper
    raise NotImplementedError


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all Virtual Networks under the same subscription

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.network.virtual_networks
    """
    result = {}
    ret_list = await hub.exec.azure.network.virtual_networks.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(
            f"Could not describe network virtual_networks {ret_list['comment']}"
        )
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.network.virtual_networks.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """Default implemented for each module."""
    return hub.tool.azure.resource_utils.is_pending(
        ret=ret, state=state, **pending_kwargs
    )



import copy
from typing import Any
from typing import Dict
from typing import List
from typing import Set


async def run_present(
    hub,
    ctx,
    resource_idem_path: str,
    name: str,
    resource_id: str,
    path_properties: Dict,
    query_properties: Dict,
    body_properties: Dict,
) -> Dict[str, Any]:
    wrapper_result = ctx.get("wrapper_result")
    computed = ctx.get("computed", {})
    computed_resource_id = computed.get("resource_id")
    computed_resource_url = computed.get("resource_url")

    result = wrapper_result

    if ctx.get("skip_present"):
        return result

    if not wrapper_result or not computed_resource_id or not computed_resource_url:
        error_message = (
            hub.tool.azure.comment_utils.invalid_present_wrapper_result_comment(
                resource_idem_path, name
            )
        )
        hub.log.error(f"{error_message}: {str(wrapper_result)} {str(computed)}")
        return {
            "result": False,
            "old_state": None,
            "new_state": None,
            "name": name,
            "comment": [error_message],
        }

    existing_resource = result["old_state"]

    path_properties = {k: v for k, v in path_properties.items() if v is not None}
    query_properties = {k: v for k, v in query_properties.items() if v is not None}
    body_properties = {k: v for k, v in body_properties.items() if v is not None}

    if not existing_resource:
        if ctx.get("test", False):
            # Return a proposed state by Idem state --test
            result["new_state"] = hub.tool.azure.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": computed_resource_id,
                    **path_properties,
                    **query_properties,
                    **body_properties,
                },
            )

            result["comment"].append(
                hub.tool.azure.comment_utils.would_create_comment(
                    resource_idem_path, name
                )
            )
            return result
        else:
            # PUT operation to create a resource
            # logic in the wrapper has made sure that create operation is allowed
            raw_body_payload = getattr(
                hub.tool, f"{resource_idem_path}.convert_present_to_raw_state"
            )(body_properties)

            request_result = await hub.tool.azure.generic.run_put_request(
                ctx,
                resource_id=computed_resource_id,
                resource_url=computed_resource_url,
                path_properties=path_properties,
                query_properties=query_properties,
                raw_payload=raw_body_payload,
                resource_idem_path=resource_idem_path,
                name=name,
                mode="create",
            )
            result["result"] = request_result["result"]
            result["comment"].extend(request_result["comment"])
            result["new_state"] = request_result["new_state"]
            return result

    else:
        present_update_payload = hub.tool.azure.generic.get_update_payload(
            existing_resource, body_properties, path_properties, query_properties
        )

        if not present_update_payload:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                hub.tool.azure.comment_utils.no_property_to_be_updated_comment(
                    resource_idem_path, name
                )
            )
            return result

        if ctx.get("test", False):
            # Return a proposed state by Idem state --test
            result["new_state"] = hub.tool.azure.test_state_utils.generate_test_state(
                enforced_state=existing_resource,
                desired_state={
                    "name": name,
                    "resource_id": computed_resource_id,
                    **path_properties,
                    **query_properties,
                    **body_properties,
                },
            )

            result["comment"].append(
                hub.tool.azure.comment_utils.would_update_comment(
                    resource_idem_path, name
                )
            )
            return result

        # PUT operation to update a resource
        raw_update_payload = getattr(
            hub.tool, f"{resource_idem_path}.convert_present_to_raw_state"
        )(present_update_payload)

        request_result = await hub.tool.azure.generic.run_put_request(
            ctx,
            resource_id=computed_resource_id,
            resource_url=computed_resource_url,
            path_properties=path_properties,
            query_properties=query_properties,
            raw_payload=raw_update_payload,
            resource_idem_path=resource_idem_path,
            name=name,
            mode="update",
        )

        result["result"] = request_result["result"]
        result["comment"].extend(request_result["comment"])
        result["new_state"] = request_result["new_state"]

        return result


async def run_put_request(
    hub,
    ctx,
    resource_id: str,
    resource_url: str,
    path_properties: str,
    query_properties: Dict,
    raw_payload: Dict,
    resource_idem_path: str,
    name: str,
    mode: str,
) -> Dict[str, Any]:
    fail_comment = ""
    success_comment = ""
    if mode == "create":
        fail_comment = hub.tool.azure.comment_utils.could_not_create_comment(
            resource_idem_path, name
        )
        success_comment = hub.tool.azure.comment_utils.create_comment(
            resource_idem_path, name
        )
    elif mode == "update":
        fail_comment = hub.tool.azure.comment_utils.could_not_update_comment(
            resource_idem_path, name
        )
        success_comment = hub.tool.azure.comment_utils.update_comment(
            resource_idem_path, name
        )

    result = await _run_json_request(
        hub,
        ctx,
        url=resource_url,
        request_method="put",
        resource_idem_path=resource_idem_path,
        success_codes=[200, 201, 202],
        success_comment=success_comment,
        fail_comment=fail_comment,
        query_params=query_properties,
        raw_payload=raw_payload,
        resource_url=resource_url,
    )

    if result["result"]:
        result["new_state"].update(path_properties)
        result["new_state"].update(query_properties)
        result["new_state"].update({"name": name, "resource_id": resource_id})

    return result


def compute_update_payload_for_key_subset(
    hub, old_values: Dict[str, Any], new_values: Dict[str, Any], key_subset: Set[str]
) -> Dict[str, Any]:
    old_values_generic = {k: v for (k, v) in old_values.items() if k in key_subset}
    new_values_generic = {k: v for (k, v) in new_values.items() if k in key_subset}
    update_payload = hub.tool.azure.generic.get_update_payload(
        old_values_generic, new_values_generic
    )
    return update_payload


def get_update_payload(
    hub,
    old_state: Dict,
    new_body_properties: Dict,
    path_properties: Dict = None,
    query_properties: Dict = None,
) -> Dict[str, Any]:
    # Exclude the path and query properties from the diff
    old_body_properties = copy.deepcopy(old_state)
    for prop_name in old_state:
        if (path_properties and prop_name in path_properties) or (
            query_properties and prop_name in query_properties
        ):
            old_body_properties.pop(prop_name, None)

    has_change = hub.tool.azure.compare.compare_exact_matches(
        old_state=old_body_properties,
        plan_state=new_body_properties,
    )

    if has_change:
        cleaned_up_new_body_props = hub.tool.azure.utils.cleanup_none_values(
            new_body_properties
        )
        changed_props_payload = copy.deepcopy(cleaned_up_new_body_props)
        for old_body_prop_name, old_body_prop_value in old_body_properties.items():
            if old_body_prop_name not in changed_props_payload:
                changed_props_payload[old_body_prop_name] = old_body_prop_value

        return changed_props_payload

    return {}


def convert_state_format(
    hub, state: Any, formatting_map: Any, list_items_indexes: Dict = None
) -> Any:
    """
    Converts a state from one format to another, e.g. present (snake case) to raw (camel case) format or vice versa.
    The conversion can change the property names and the structure of the resulting state, e.g. flatten
    deeply nested properties. Works with nested lists and dicts.

    Args:
        hub: The redistributed pop central hub.
        state: the current state of a resource.
        formatting_map: a mapping from one format to another. E.g.:
            "virtualNetworkPeerings": [
                {
                    "$list_refs": ["virtual_network_peerings"],
                    "name": "virtual_network_peerings.name",
                    "type": "virtual_network_peerings.type",
                    "properties": {
                        "allowVirtualNetworkAccess": "virtual_network_peerings.allow_virtual_network_access",
                        "allowForwardedTraffic": "virtual_network_peerings.allow_forwarded_traffic",
                        "allowGatewayTransit": "virtual_network_peerings.allow_gateway_transit",
                        "useRemoteGateways": "virtual_network_peerings.use_remote_gateways",
                        "remoteVirtualNetwork": "virtual_network_peerings.remote_virtual_network",
                        "remoteAddressSpace": {
                            "addressPrefixes": "virtual_network_peerings.remote_address_space.address_prefixes"
                        },
                        "remoteVirtualNetworkAddressSpace": {
                            "addressPrefixes": "virtual_network_peerings.remote_virtual_network_address_space.address_prefixes"
                        },
                        "remoteBgpCommunities": {
                            "virtualNetworkCommunity": "virtual_network_peerings.remote_bgp_communities.virtual_network_community"
                        },
                        "peeringState": "virtual_network_peerings.peering_state",
                        "peeringSyncLevel": "virtual_network_peerings.peering_sync_level",
                        "doNotVerifyRemoteGateways": "virtual_network_peerings.do_not_verify_remote_gateways",
                    },
                }
            ]
            $list_refs is a special reserved property in the mapping. Use it to specify which lists from the original
            state should be tracked at the current nesting level
        list_items_indexes: ignore, used internally by this recursive method.
            This property keeps track of the nested lists indexes

    Returns:
        The state in the new format
    """
    formatted_state = {}

    if isinstance(formatting_map, dict):
        for formatted_name, formatting_map_for_item in formatting_map.items():
            if formatted_name == "$list_refs":
                continue

            converted_value = hub.tool.azure.generic.convert_state_format(
                state, formatting_map_for_item, list_items_indexes
            )
            if (
                isinstance(formatting_map_for_item, str) and converted_value is not None
            ) or (
                isinstance(formatting_map_for_item, (dict, list)) and converted_value
            ):
                formatted_state[formatted_name] = converted_value
        return formatted_state
    elif isinstance(formatting_map, list):
        ref_lists_names = formatting_map[0]["$list_refs"]
        if not isinstance(ref_lists_names, list):
            ref_lists_names = [ref_lists_names]
        ref_lists = [
            _get_element_from_state(state, ref_list_name, list_items_indexes)
            for ref_list_name in ref_lists_names
        ]
        ref_lists = [el for el in ref_lists if el is not None]
        longest_ref_list = max(ref_lists, key=len) if ref_lists else []
        formatted_list = []
        for i, item in enumerate(longest_ref_list):
            if list_items_indexes is None:
                list_items_indexes = {}
            for ref_list_name in ref_lists_names:
                list_items_indexes[ref_list_name] = i
            formatted_list.append(
                hub.tool.azure.generic.convert_state_format(
                    state, formatting_map[0], list_items_indexes
                )
            )

        if list_items_indexes:
            for ref_list_name in ref_lists_names:
                list_items_indexes.pop(ref_list_name, None)

        return formatted_list
    elif isinstance(formatting_map, str):
        return _get_element_from_state(state, formatting_map, list_items_indexes)

    return formatted_state


def _get_element_from_state(
    state: Any, path: str, list_item_indexes: Dict = None
) -> Any:
    path_segments = path.split(".")
    partial_path = None
    element = state
    for path_segment in path_segments:
        if element.get(path_segment) is None:
            return None
        element = element[path_segment]
        partial_path = (
            f"{partial_path}.{path_segment}" if partial_path else path_segment
        )
        if list_item_indexes and partial_path in list_item_indexes:
            i = list_item_indexes[partial_path]
            if i < len(element):
                element = element[i]
            else:
                return None

    return element


async def _run_json_request(
    hub,
    ctx,
    url: str,
    request_method: str,
    resource_idem_path: str,
    success_comment: str,
    fail_comment: str,
    success_codes: List[int],
    query_params: Dict = None,
    raw_payload: Dict = None,
    resource_url: str = None,
):
    result = {"comment": [], "result": True, "new_state": None}

    if query_params:
        query_params_url = "&".join(
            [
                f"{prop_name}={prop_val}"
                for prop_name, prop_val in query_params.items()
                if prop_val is not None
            ]
        )
        if query_params_url:
            url = f"{url}&{query_params_url}"

    request_body = {
        "url": url,
        "success_codes": success_codes,
    }

    if raw_payload:
        request_body.update(
            {
                "json": raw_payload,
            }
        )

    response = await getattr(hub.exec.request.json, request_method)(
        ctx,
        **request_body,
    )

    if response["status"] == 202:
        response = await hub.tool.azure.operation_utils.await_operation(
            ctx,
            operation_headers=response.get("headers"),
            resource_url=resource_url,
        )

    if not response["result"]:
        result["result"] = False
        result["comment"].append(fail_comment)
        if response.get("comment"):
            result["comment"].append(response["comment"])
        if response.get("ret"):
            result["comment"].append(str(response["ret"]))
        hub.log.debug(f"{result['comment']}")
        return result

    if response.get("ret"):
        result["new_state"] = getattr(
            hub.tool, f"{resource_idem_path}.convert_raw_to_present_state"
        )(response["ret"])

    result["comment"].append(success_comment)
    return result


async def exec_request(
    hub,
    ctx,
    resource_id: str,
    request_method: str,
    api_method: str,
    api_version: str,
    resource_idem_path: str,
    query_params: Dict = None,
    raw_payload: Dict = None,
) -> Dict[str, Any]:
    success_comment = hub.tool.azure.comment_utils.executed_request_comment(
        resource_idem_path, resource_id, api_method
    )
    fail_comment = hub.tool.azure.comment_utils.could_not_execute_request_comment(
        resource_idem_path, resource_id, api_method
    )
    api_method_url = (
        f"{ctx.acct.endpoint_url}{resource_id}/{api_method}?api-version={api_version}"
    )
    resource_url = f"{ctx.acct.endpoint_url}{resource_id}?api-version={api_version}"
    return await _run_json_request(
        hub,
        ctx,
        url=api_method_url,
        request_method=request_method,
        resource_idem_path=resource_idem_path,
        success_codes=[200],
        success_comment=success_comment,
        fail_comment=fail_comment,
        query_params=query_params,
        raw_payload=raw_payload,
        resource_url=resource_url,
    )



"""State module for managing Storage Accounts."""
import copy
import uuid
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    account_name: str,
    location: str,
    sku_name: str,
    sku_tier: str = None,
    account_kind: str = None,
    cross_tenant_replication_enabled: bool = None,
    access_tier: str = None,
    edge_zone: str = None,
    enable_https_traffic_only: bool = None,
    min_tls_version: str = None,
    allow_blob_public_access: bool = None,
    allow_shared_key_access: bool = None,
    public_network_access: str = None,
    default_to_oauth_authentication: bool = None,
    is_hns_enabled: bool = None,
    nfsv3_enabled: bool = None,
    custom_domain: make_dataclass(
        "CustomDomain",
        [
            ("name", str),
            ("use_subdomain", bool, field(default=None)),
        ],
    ) = None,
    customer_managed_key: make_dataclass(
        "CustomerManagedKey",
        [
            ("user_assigned_identity_id", str),
            ("key_vault_key_id", str, field(default=None)),
            ("federated_identity_client_id", str, field(default=None)),
            ("key_name", str, field(default=None)),
            ("key_vault_uri", str, field(default=None)),
            ("key_version", str, field(default=None)),
        ],
    ) = None,
    identity: make_dataclass(
        "Identity",
        [
            ("type", str),
            ("user_assigned_identities", Dict[str, str], field(default=None)),
        ],
    ) = None,
    network_rules: make_dataclass(
        "NetworkRules",
        [
            ("default_action", str),
            ("bypass", str, field(default=None)),
            ("ip_rule_values", List[str], field(default=None)),
            ("virtual_network_subnet_ids", List[str], field(default=None)),
            (
                "resource_access_rules",
                List[
                    make_dataclass(
                        "ResourceAccessRules",
                        [
                            ("endpoint_resource_id", str),
                            ("endpoint_tenant_id", str, field(default=None)),
                        ],
                    )
                ],
                field(default=None),
            ),
        ],
    ) = None,
    large_file_shares_state: str = None,
    azure_files_authentication: make_dataclass(
        "AzureFilesAuthentication",
        [
            ("directory_service_options", str),
            (
                "active_directory_properties",
                make_dataclass(
                    "ActiveDirectoryProperties",
                    [
                        ("azure_storage_sid", str),
                        ("domain_name", str),
                        ("domain_sid", str),
                        ("domain_guid", str),
                        ("forest_name", str),
                        ("netbios_domain_name", str),
                    ],
                ),
                field(default=None),
            ),
            ("default_share_permission", str, field(default=None)),
        ],
    ) = None,
    routing: make_dataclass(
        "Routing",
        [
            ("publish_internet_endpoints", bool, field(default=None)),
            ("publish_microsoft_endpoints", bool, field(default=None)),
            ("routing_choice", str, field(default=None)),
        ],
    ) = None,
    encryption_service: make_dataclass(
        "EncryptionService",
        [
            ("encryption_key_source", str, field(default="Microsoft.Storage")),
            ("queue_encryption_key_type", str, field(default=None)),
            ("table_encryption_key_type", str, field(default=None)),
            ("blob_encryption_key_type", str, field(default=None)),
            ("file_encryption_key_type", str, field(default=None)),
        ],
    ) = None,
    require_infrastructure_encryption: bool = None,
    immutability_policy: make_dataclass(
        "ImmutabilityPolicy",
        [
            ("allow_protected_append_writes", bool),
            ("state", str),
            ("period_since_creation_in_days", int),
        ],
    ) = None,
    sas_policy: make_dataclass(
        "SasPolicy",
        [
            ("expiration_period", str),
            ("expiration_action", str, field(default="Log")),
        ],
    ) = None,
    key_policy: make_dataclass(
        "KeyPolicy",
        [
            ("key_expiration_period_in_days", int),
        ],
    ) = None,
    allowed_copy_scope: str = None,
    sftp_enabled: bool = None,
    tags: Dict[str, str] = None,
    subscription_id: str = None,
    resource_id: str = None,
) -> Dict:
    r"""Create or update Storage Accounts.

    Args:
        name(str): The identifier for this state.
        account_name(str): The name of the storage account within the specified resource group. Storage account names
            must be between 3 and 24 characters in length and use numbers and lower-case letters only.
        resource_group_name(str): The name of the resource group.
        location(str): Specifies the supported Azure location where the resource exists.
        sku_name(str): The SKU name.
        sku_tier(str, Optional): The SKU tier.
        account_kind(str, Optional): Defines the Kind of account.
        cross_tenant_replication_enabled(bool, Optional): Allow or disallow cross AAD tenant object replication
        access_tier(str, Optional): Required for storage accounts where account_kind = BlobStorage. The access tier is used for billing.
        edge_zone(str, Optional): Specifies the Edge Zone within the Azure Region where this Storage Account should exist.
        enable_https_traffic_only(bool, Optional): Boolean flag which forces HTTPS if enabled.
        min_tls_version(str, Optional): The minimum supported TLS version for the storage account.
        allow_blob_public_access(bool, Optional): Allow or disallow nested items within this Account to opt into being public.
        allow_shared_key_access(bool, Optional): Indicates whether the storage account permits requests to be authorized with the account access key via Shared Key.
        public_network_access(str, Optional): Allow or disallow public network access to Storage Account. Value is optional but if passed in, must be 'Enabled' or 'Disabled'.
        default_to_oauth_authentication(bool, Optional): A boolean flag which indicates whether the default authentication is OAuth or not. The default interpretation is false for this property.
        is_hns_enabled(bool, Optional): Account HierarchicalNamespace enabled if sets to true.
        nfsv3_enabled(bool, Optional): NFS 3.0 protocol support enabled if set to true.
        custom_domain(dict[str, Any], Optional): User domain assigned to the storage account.

            * name(str):
                The Custom Domain Name to use for the Storage Account.
            * use_subdomain(bool, Optional):
                Indicates whether indirect CName validation is enabled. Default value is false.
        customer_managed_key(dict[str, Any], Optional): Combination of Key vault key id and user assigned id.

            * user_assigned_identity_id(str):
                User Assigned Identity id.
            * key_vault_key_id(str, Optional):
                The object identifier of the current versioned Key Vault Key in use.
            * federated_identity_client_id(str, Optional):
                ClientId of the multi-tenant application to be used in conjunction with the user-assigned identity for cross-tenant customer-managed-keys server-side encryption on the storage account.
            * key_name(str, Optional):
                The name of KeyVault key.
            * key_vault_uri(str, Optional):
                The Uri of KeyVault.
            * key_version(str, Optional):
               The version of KeyVault key.
        identity(dict[str, Any], Optional): The identity of the resource.

            * type(str):
                The identity type. Possible values are SystemAssigned, UserAssigned, SystemAssigned, UserAssigned.
            * user_assigned_identities(dict[str, Any], Optional):
                Key value pairs that describe the set of User Assigned identities that will be used with this storage account.
        network_rules(dict[str, Any], Optional): Network rule set.

            * default_action(str):
                Specifies the default action of allow or deny when no other rules match. Valid options are Deny or Allow.
            * bypass(str, Optional):
                Specifies whether traffic is bypassed for Logging/Metrics/AzureServices. Possible values are any combination of Logging|Metrics|AzureServices.
            * ip_rule_values(list, Optional):
                List of IP or IP range in CIDR format. Only IPV4 address is allowed.
            * virtual_network_subnet_ids(list, Optional):
                A list of resource ids of virtual network subnets.
            * resource_access_rules(dict[str, Any], Optional):
                The resource access rules.
        large_file_shares_state(str, Optional): Allow large file shares if sets to enable.
        azure_files_authentication(dict[str, Any], Optional): Provides the identity based authentication settings for Azure Files.

            * directory_service_options(str):
                The directory service to be used. Possible values are AADDS, AD and AADKERB.
            * active_directory_properties(dict[str, Any], Optional):
                Required if directoryServiceOptions are AD, optional if they are AADKERB.

                * azure_storage_sid(str):
                    Specifies the security identifier (SID) for Azure Storage.
                * domain_name(str):
                    Specifies the primary domain that the AD DNS server is authoritative for.
                * domain_sid(str):
                    Specifies the security identifier (SID).
                * domain_guid(str):
                    Specifies the domain GUID.
                * forest_name(str):
                   Specifies the Active Directory forest to get.
                * netbios_domain_name(str):
                    Specifies the NetBIOS domain name.
            * default_share_permission(str):
                Default share permission for users using Kerberos authentication if RBAC role is not assigned.
        routing(dict[str, Any], Optional): Maintains information about the network routing choice opted by the user for data transfer.

            * publish_internet_endpoints(bool, Optional):
                A boolean flag which indicates whether internet routing storage endpoints are to be published.
            * publish_microsoft_endpoints(bool, Optional):
                A boolean flag which indicates whether microsoft routing storage endpoints are to be published.
            * routing_choice(str, Optional):
                Routing Choice defines the kind of network routing opted by the user.
        encryption_service(dict[str, Any], Optional): Encryption details.

            * queue_encryption_key_type(str, Optional):
                The encryption type of the queue service.
            * table_encryption_key_type(str, Optional):
                The encryption type of the table service.
            * blob_encryption_key_type(str, Optional):
                The encryption type of the blob service.
            * file_encryption_key_type(str, Optional):
                The encryption type of the file service.
            * encryption_key_source(str, Optional):
                The encryption keySource (provider)
        require_infrastructure_encryption(bool, Optional): A boolean indicating whether or not the service applies a secondary layer of encryption with platform managed keys for data at rest.
        immutability_policy(dict[str, Any], Optional): This argument specifies the default account-level immutability policy which is inherited and applied to objects.

            * allow_protected_append_writes(bool):
                This property can only be changed for disabled and unlocked time-based retention policies. When enabled, new blocks can be written to an append blob while
                maintaining immutability protection and compliance. Only new blocks can be added and any existing blocks cannot be modified or deleted.
            * state(str):
                Defines the mode of the policy. Disabled state disables the policy, Unlocked state
                allows increase and decrease of immutability retention time and also allows toggling allowProtectedAppendWrites property,
                Locked state only allows the increase of the immutability retention time. A policy can only be created
                in a Disabled or Unlocked state and can be toggled between the two states. Only a policy in an Unlocked
                state can transition to a Locked state which cannot be reverted.
            * period_since_creation_in_days(int): The immutability period for the blobs in the container since the policy creation, in days.
        sas_policy(dict[str, Any], Optional): SasPolicy assigned to the storage account.

            * expiration_period(str, Optional):
                The SAS expiration period, DD.HH:MM:SS.
            * expiration_action(str, Optional):
                The SAS expiration action. Valid value is Log.
        key_policy(dict[str, Any], Optional): KeyPolicy assigned to the storage account.

            * key_expiration_period_in_days(int, Optional):
                The key expiration period in days.
        allowed_copy_scope(str, Optional): Restrict copy to and from Storage Accounts within an AAD tenant or with Private Links to the same VNet.
        sftp_enabled(bool, Optional): Enables Secure File Transfer Protocol, if set to true.
        tags(dict[str, str], Optional): The resource tags.
        subscription_id(str, Optional): Subscription Unique id.
        resource_id(str, Optional): Storage account resource id on Azure.

    Returns:
        dict

    Examples:
        .. code-block:: sls

            resource_is_present:
              azure.storage_resource_provider.storage_accounts.present:
                - resource_group_name: value
                - account_name: value
                - location: value
                - sku_name: value
                - sku_tier: value
    """

    result = ctx.get("wrapper_result")

    if ctx.get("skip_present"):
        return result

    if not result:
        error_message = hub.tool.azure.comment_utils.no_result_from_wrapper(
            "azure.storage_resource_provider.storage_accounts", name
        )
        hub.log.error(error_message)
        return {
            "result": False,
            "old_state": None,
            "new_state": None,
            "name": name,
            "comment": [error_message],
        }

    computed = ctx.get("computed")
    computed_resource_id = computed.get("resource_id")
    computed_resource_url = computed.get("resource_url")
    existing_resource = result["old_state"]

    if not existing_resource:
        if ctx.get("test", False):
            # Return a proposed state by Idem state --test
            result["new_state"] = hub.tool.azure.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_group_name": resource_group_name,
                    "account_name": account_name,
                    "location": location,
                    "sku_tier": sku_tier,
                    "sku_name": sku_name,
                    "account_kind": account_kind,
                    "cross_tenant_replication_enabled": cross_tenant_replication_enabled,
                    "access_tier": access_tier,
                    "edge_zone": edge_zone,
                    "enable_https_traffic_only": enable_https_traffic_only,
                    "min_tls_version": min_tls_version,
                    "allow_blob_public_access": allow_blob_public_access,
                    "allow_shared_key_access": allow_shared_key_access,
                    "public_network_access": public_network_access,
                    "default_to_oauth_authentication": default_to_oauth_authentication,
                    "is_hns_enabled": is_hns_enabled,
                    "nfsv3_enabled": nfsv3_enabled,
                    "custom_domain": custom_domain,
                    "customer_managed_key": customer_managed_key,
                    "identity": identity,
                    "network_rules": network_rules,
                    "large_file_shares_state": large_file_shares_state,
                    "azure_files_authentication": azure_files_authentication,
                    "routing": routing,
                    "encryption_service": encryption_service,
                    "require_infrastructure_encryption": require_infrastructure_encryption,
                    "immutability_policy": immutability_policy,
                    "sas_policy": sas_policy,
                    "key_policy": key_policy,
                    "allowed_copy_scope": allowed_copy_scope,
                    "sftp_enabled": sftp_enabled,
                    "tags": tags,
                    "resource_id": computed_resource_id,
                    "subscription_id": subscription_id,
                },
            )
            result["comment"].append(
                f"Would create azure.storage_resource_provider.storage_accounts '{name}'"
            )
            return result
        else:
            # PUT operation to create a resource
            payload = hub.tool.azure.storage_resource_provider.storage_accounts.convert_present_to_raw_storage_accounts(
                location=location,
                sku_tier=sku_tier,
                sku_name=sku_name,
                account_kind=account_kind,
                cross_tenant_replication_enabled=cross_tenant_replication_enabled,
                access_tier=access_tier,
                edge_zone=edge_zone,
                enable_https_traffic_only=enable_https_traffic_only,
                min_tls_version=min_tls_version,
                allow_blob_public_access=allow_blob_public_access,
                allow_shared_key_access=allow_shared_key_access,
                public_network_access=public_network_access,
                default_to_oauth_authentication=default_to_oauth_authentication,
                is_hns_enabled=is_hns_enabled,
                nfsv3_enabled=nfsv3_enabled,
                custom_domain=custom_domain,
                customer_managed_key=customer_managed_key,
                identity=identity,
                network_rules=network_rules,
                large_file_shares_state=large_file_shares_state,
                azure_files_authentication=azure_files_authentication,
                routing=routing,
                encryption_service=encryption_service,
                require_infrastructure_encryption=require_infrastructure_encryption,
                immutability_policy=immutability_policy,
                sas_policy=sas_policy,
                key_policy=key_policy,
                allowed_copy_scope=allowed_copy_scope,
                sftp_enabled=sftp_enabled,
                tags=tags,
            )
            response_put = await hub.exec.request.json.put(
                ctx,
                url=computed_resource_url,
                success_codes=[200, 201],
                json=payload,
            )

            if response_put["status"] == 202:
                result["rerun_data"] = {
                    "operation_id": str(uuid.uuid4()),
                    "operation_headers": dict(response_put.get("headers")),
                    "resource_id": f"{computed_resource_id}",
                    "resource_url": f"{computed_resource_url}",
                    "old_state": result["old_state"],
                }
                return result

            elif not response_put["result"] and response_put["status"] != 202:
                hub.log.debug(
                    f"Could not create azure.storage_resource_provider.storage_accounts {response_put['comment']} {response_put['ret']}"
                )
                result["comment"].extend(
                    hub.tool.azure.result_utils.extract_error_comments(response_put)
                )
                result["result"] = False
                return result

            else:
                result[
                    "new_state"
                ] = hub.tool.azure.storage_resource_provider.storage_accounts.convert_raw_storage_accounts_to_present(
                    resource=response_put["ret"],
                    idem_resource_name=name,
                    resource_group_name=resource_group_name,
                    account_name=account_name,
                    resource_id=computed_resource_id,
                    subscription_id=subscription_id,
                )

            result["comment"].append(
                f"Created azure.storage_resource_provider.storage_accounts '{name}'"
            )
            return result

    else:
        raw_existing_resource = hub.tool.azure.storage_resource_provider.storage_accounts.convert_present_to_raw_storage_accounts(
            **existing_resource
        )
        # Generate a new PUT operation payload with new values
        new_payload = hub.tool.azure.storage_resource_provider.storage_accounts.update_storage_accounts_payload(
            raw_existing_resource,
            {
                "sku_tier": sku_tier,
                "sku_name": sku_name,
                "account_kind": account_kind,
                "cross_tenant_replication_enabled": cross_tenant_replication_enabled,
                "access_tier": access_tier,
                "edge_zone": edge_zone,
                "enable_https_traffic_only": enable_https_traffic_only,
                "min_tls_version": min_tls_version,
                "allow_blob_public_access": allow_blob_public_access,
                "allow_shared_key_access": allow_shared_key_access,
                "public_network_access": public_network_access,
                "default_to_oauth_authentication": default_to_oauth_authentication,
                "is_hns_enabled": is_hns_enabled,
                "nfsv3_enabled": nfsv3_enabled,
                "custom_domain": custom_domain,
                "customer_managed_key": customer_managed_key,
                "identity": identity,
                "network_rules": network_rules,
                "large_file_shares_state": large_file_shares_state,
                "azure_files_authentication": azure_files_authentication,
                "routing": routing,
                "encryption_service": encryption_service,
                # "queue_encryption_key_type": queue_encryption_key_type,
                # "table_encryption_key_type": table_encryption_key_type,
                # "encryption_key_source": encryption_key_source,
                "require_infrastructure_encryption": require_infrastructure_encryption,
                "immutability_policy": immutability_policy,
                "sas_policy": sas_policy,
                "key_policy": key_policy,
                "allowed_copy_scope": allowed_copy_scope,
                "sftp_enabled": sftp_enabled,
                "tags": tags,
            },
        )

        if ctx.get("test", False):
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    hub.tool.azure.comment_utils.no_property_to_be_updated_comment(
                        "azure.storage_resource_provider.storage_accounts", name
                    )
                )
            else:
                result[
                    "new_state"
                ] = hub.tool.azure.storage_resource_provider.storage_accounts.convert_raw_storage_accounts_to_present(
                    resource=new_payload["ret"],
                    idem_resource_name=name,
                    resource_group_name=resource_group_name,
                    account_name=account_name,
                    resource_id=computed_resource_id,
                    subscription_id=subscription_id,
                )
                result["comment"].append(
                    f"Would update azure.storage_resource_provider.storage_accounts '{name}'"
                )
            return result

        # PUT operation to update a resource
        if new_payload["ret"] is None:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                hub.tool.azure.comment_utils.no_property_to_be_updated_comment(
                    "azure.storage_resource_provider.storage_accounts", name
                )
            )
            return result
        result["comment"].extend(new_payload["comment"])
        response_put = await hub.exec.request.json.put(
            ctx,
            url=computed_resource_url,
            success_codes=[200],
            json=new_payload["ret"],
        )
        if response_put["status"] == 202:
            result["rerun_data"] = {
                "operation_id": str(uuid.uuid4()),
                "operation_headers": dict(response_put.get("headers")),
                "resource_id": f"{computed_resource_id}",
                "resource_url": f"{computed_resource_url}",
                "old_state": result["old_state"],
            }
            return result
        elif not response_put["result"] and response_put["status"] != 202:
            hub.log.debug(
                f"Could not update azure.storage_resource_provider.storage_accounts {response_put['comment']} {response_put['ret']}"
            )
            result["result"] = False
            result["comment"].extend(
                hub.tool.azure.result_utils.extract_error_comments(response_put)
            )
            return result

        result[
            "new_state"
        ] = hub.tool.azure.storage_resource_provider.storage_accounts.convert_raw_storage_accounts_to_present(
            resource=response_put["ret"],
            idem_resource_name=name,
            resource_group_name=resource_group_name,
            account_name=account_name,
            resource_id=computed_resource_id,
            subscription_id=subscription_id,
        )
        if result["old_state"] == result["new_state"]:
            result["comment"].append(
                hub.tool.azure.comment_utils.no_property_to_be_updated_comment(
                    "azure.storage_resource_provider.storage_accounts", name
                )
            )
            return result
        result["comment"].append(
            f"Updated azure.storage_resource_provider.storage_accounts '{name}'"
        )
        return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    resource_group_name: str = None,
    account_name: str = None,
    subscription_id: str = None,
) -> dict:
    r"""Delete Storage Accounts.

    Args:
        name(str): The identifier for this state.
        resource_id(str, Optional): Storage Account resource id on Azure.
        resource_group_name(str): The name of the resource group.
        account_name(str): The name of the storage account within the specified resource group.
        subscription_id(str, Optional): Subscription Unique id.

    Returns:
        dict

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.storage_resource_provider.storage_accounts.absent:
                - name: value
                - resource_group_name: value
                - account_name: value
    """
    return hub.tool.azure.result_utils.absent_implemented_through_wrapper_result(
        "azure.storage_resource_provider.storage_accounts", name
    )


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all Storage Accounts under the same subscription.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.storage_resource_provider.storage_accounts
    """
    result = {}
    ret_list = await hub.exec.azure.storage_resource_provider.storage_accounts.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(f"Could not describe storage account {ret_list['comment']}")
        return result
    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.storage_resource_provider.storage_accounts.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """Default implemented for each module."""
    return hub.tool.azure.resource_utils.is_pending(
        ret=ret, state=state, **pending_kwargs
    )



"""State module for managing Policy Assignments."""
import copy
from typing import Any
from typing import Dict


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    scope: str,
    policy_assignment_name: str,
    policy_definition_id: str,
    parameters: Dict[str, Any] = None,
    resource_id: str = None,
) -> Dict:
    r"""Create or update Policy Assignments.

    Args:
        name(str): The identifier for this state.
        scope(str): The scope of the policy assignment. Valid scopes are: management group (format: '/providers/Microsoft.Management/managementGroups/{managementGroup}'), subscription (format: '/subscriptions/{subscriptionId}'), resource group (format: '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}', or resource (format: '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{resourceProviderNamespace}/[{parentResourcePath}/]{resourceType}/{resourceName}'.
        policy_assignment_name(str): The name which should be used for this Policy Assignment. Changing this forces a new Resource Policy Assignment to be created.
        policy_definition_id(str): The ID of the Policy Definition or Policy Definition Set. Changing this forces a new Policy Assignment to be created
        parameters(dict[str, Any], Optional): Policy assignment parameters with respect to policy definition rule. Defaults to {}.
        resource_id(str, Optional): Policy Assignment resource id on Azure

    Returns:
        dict

    Examples:
        .. code-block:: sls

            resource_is_present:
              azure.policy.policy_assignments.present:
                - name: value
                - scope: value
                - policy_assignment_name: value
                - policy_definition_id: value
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if parameters is None:
        parameters = {}

    if resource_id is None:
        resource_id = f"{scope}/providers/Microsoft.Authorization/policyAssignments/{policy_assignment_name}"

    response_get = await hub.exec.azure.policy.policy_assignments.get(
        ctx, resource_id=resource_id, raw=True
    )

    if response_get["result"]:
        if not response_get["ret"]:
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "policy_assignment_name": policy_assignment_name,
                        "policy_definition_id": policy_definition_id,
                        "scope": scope,
                        "parameters": parameters,
                        "resource_id": resource_id,
                    },
                )
                result["comment"].append(
                    f"Would create azure.policy.policy_assignments '{name}'"
                )
                return result
            else:
                # PUT operation to create a resource
                payload = hub.tool.azure.policy.policy_assignment.convert_present_to_raw_policy_assignment(
                    policy_definition_id=policy_definition_id,
                    parameters=parameters,
                )
                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
                    success_codes=[200, 201],
                    json=payload,
                )
                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not create azure.policy.policy_assignments {response_put['comment']} {response_put['ret']}"
                    )
                    result["comment"].extend(
                        hub.tool.azure.result_utils.extract_error_comments(response_put)
                    )
                    result["result"] = False
                    return result

                result[
                    "new_state"
                ] = hub.tool.azure.policy.policy_assignment.convert_raw_policy_assignment_to_present(
                    resource=response_put["ret"],
                    idem_resource_name=name,
                    resource_id=resource_id,
                )
                result["comment"].append(
                    f"Created azure.policy.policy_assignments '{name}'"
                )
                return result

        else:
            existing_resource = response_get["ret"]
            result[
                "old_state"
            ] = hub.tool.azure.policy.policy_assignment.convert_raw_policy_assignment_to_present(
                resource=existing_resource,
                idem_resource_name=name,
                resource_id=resource_id,
            )
            # Generate a new PUT operation payload with new values
            new_payload = hub.tool.azure.policy.policy_assignment.update_policy_assignment_payload(
                existing_resource,
                {
                    "policy_definition_id": policy_definition_id,
                    "parameters": parameters,
                },
            )
            if ctx.get("test", False):
                if new_payload["ret"] is None:
                    result["new_state"] = copy.deepcopy(result["old_state"])
                    result["comment"].append(
                        f"azure.policy.policy_assignments '{name}' has no property need to be updated."
                    )
                else:
                    result[
                        "new_state"
                    ] = hub.tool.azure.policy.policy_assignment.convert_raw_policy_assignment_to_present(
                        resource=new_payload["ret"],
                        idem_resource_name=name,
                        resource_id=resource_id,
                    )
                    result["comment"].append(
                        f"Would update azure.policy.policy_assignments '{name}'"
                    )
                return result

            # PUT operation to update a resource
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.policy.policy_assignments '{name}' has no property need to be updated."
                )
                return result
            result["comment"].extend(new_payload["comment"])
            response_put = await hub.exec.request.json.put(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
                success_codes=[200, 201],
                json=new_payload["ret"],
            )
            if not response_put["result"]:
                hub.log.debug(
                    f"Could not update azure.policy.policy_assignments {response_put['comment']} {response_put['ret']}"
                )
                result["result"] = False
                result["comment"].extend(
                    hub.tool.azure.result_utils.extract_error_comments(response_put)
                )
                return result

            result[
                "new_state"
            ] = hub.tool.azure.policy.policy_assignment.convert_raw_policy_assignment_to_present(
                resource=response_put["ret"],
                idem_resource_name=name,
                resource_id=resource_id,
            )
            if result["old_state"] == result["new_state"]:
                result["comment"].append(
                    f"azure.policy.policy_assignments '{name}' has no property need to be updated."
                )
                return result
            result["comment"].append(
                f"Updated azure.policy.policy_assignments '{name}'"
            )
            return result
    else:
        hub.log.debug(
            f"Could not get azure.policy.policy_assignments {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"].extend(
            hub.tool.azure.result_utils.extract_error_comments(response_get)
        )
        return result


async def absent(hub, ctx, name: str, scope: str, policy_assignment_name: str) -> dict:
    r"""Delete Policy Assignment.

    Args:
        name(str): The identifier for this state.
        scope(str): The scope of the policy assignment. Valid scopes are: management group (format: '/providers/Microsoft.Management/managementGroups/{managementGroup}'), subscription (format: '/subscriptions/{subscriptionId}'), resource group (format: '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}', or resource (format: '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{resourceProviderNamespace}/[{parentResourcePath}/]{resourceType}/{resourceName}'.
        policy_assignment_name(str): The name of the policy assignment to delete.

    Returns:
        dict

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.policy.policy_assignments.absent:
                - name: value
                - scope: value
                - policy_assignment_name: value
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    resource_id = f"{scope}/providers/Microsoft.Authorization/policyAssignments/{policy_assignment_name}"

    response_get = await hub.exec.azure.policy.policy_assignments.get(
        ctx,
        resource_id=resource_id,
    )
    if response_get["result"]:
        if response_get["ret"]:
            result["old_state"] = response_get["ret"]
            result["old_state"]["name"] = name
            if ctx.get("test", False):
                result["comment"].append(
                    f"Would delete azure.policy.policy_assignments '{name}'"
                )
                return result
            response_delete = await hub.exec.request.raw.delete(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
                success_codes=[200, 202],
            )

            if not response_delete["result"]:
                hub.log.debug(
                    f"Could not delete azure.policy.policy_assignments {response_delete['comment']} {response_delete['ret']}"
                )
                result["result"] = False
                result["comment"].extend(
                    hub.tool.azure.result_utils.extract_error_comments(response_delete)
                )
                return result

            result["comment"].append(
                f"Deleted azure.policy.policy_assignments '{name}'"
            )
            return result

        else:
            # If Azure returns 'Not Found' error, it means the resource has been absent.
            result["comment"].append(
                f"azure.policy.policy_assignments '{name}' already absent"
            )
            return result
    else:
        hub.log.debug(
            f"Could not get azure.policy.policy_assignments '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"].extend(
            hub.tool.azure.result_utils.extract_error_comments(response_get)
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all Policy Assignments under the same subscription.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.policy.policy_assignments
    """
    result = {}
    ret_list = await hub.exec.azure.policy.policy_assignments.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(f"Could not describe policy assignment {ret_list['comment']}")
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.policy.policy_assignments.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



NAME_PARAMETER = {
    "default": None,
    "doc": "The identifier for this state",
    "param_type": "str",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}

GET_REQUEST_FORMAT = r"""
    subscription_id = ctx.acct.subscription_id
    return {{ function.hardcoded.get_function }}(
        ctx,
        url=f"{{ function.hardcoded.path }}",
        success_codes=[200],
    )
"""

LIST_REQUEST_FORMAT = r"""
    ret = dict(comment="", result=True, ret={})
    subscription_id = ctx.acct.subscription_id

    async for page_result in hub.tool.azure.request.paginate(
            ctx,
            url=f"{{ function.hardcoded.path }}",
            success_codes=[200],
    ):
        resource_list = page_result.get("value", None)
        if resource_list:
            for resource in resource_list:
                ret["ret"][resource["id"]] = resource
    return ret
"""

CREATE_REQUEST_FORMAT = r"""
    subscription_id = ctx.acct.subscription_id

    # PUT operation to create a resource
    return {{ function.hardcoded.create_function }}(
        ctx,
        url=f"{{ function.hardcoded.path }}",
        success_codes=[200, 201],
        json=parameters,
    )
"""

UPDATE_REQUEST_FORMAT = r"""
    subscription_id = ctx.acct.subscription_id

    {% if function.hardcoded.patch_parameters %}
    response_get = {{ function.hardcoded.get_function }}(
        ctx,
        url=f"{{ function.hardcoded.path }}",
        success_codes=[200],
    )

    # PATCH operation to update a resource
    patch_parameters = {{ function.hardcoded.patch_parameters }}
    existing_resource = response_get["ret"]
    new_parameters = hub.tool.azure.request.patch_json_content(patch_parameters, existing_resource, parameters)

    if force_update:
        return {{ function.hardcoded.create_function }}(
            ctx,
            url=f"{{ function.hardcoded.path }}",
            success_codes=[200, 201],
            json=new_parameters,
        )
    else:
        return {{ function.hardcoded.patch_function }}(
            ctx,
            url=f"{{ function.hardcoded.path }}",
            success_codes=[200],
            json=new_parameters,
        )

    {% else %}
    ret = dict(
        comment="No update operation on {{ function.hardcoded.resource_name }} since Azure does not have PATCH api on {{ function.hardcoded.resource_name }}",
        result=True,
        ret=None
    )

    return ret
    {% endif %}
"""

DELETE_REQUEST_FORMAT = r"""
    subscription_id = ctx.acct.subscription_id

    return {{ function.hardcoded.delete_function }}(
        ctx,
        url=f"{{ function.hardcoded.path }}",
        success_codes=[200, 202, 204],
    )
"""

PRESENT_REQUEST_FORMAT = r"""
    # TODO Please use virtual_network.py as a reference. Function input parameters need to be added manually.
    result = dict(name=name, result=True, old_state=None, new_state=None, comment=())
    subscription_id = ctx.acct.subscription_id
    if resource_id is None:
        # TODO: This resource_id is not correct and it needs to be manually fixed to the correct one
        resource_id = {{ function.hardcoded.get_function }}
    # TODO: double check if the api version here is correct or not.
    response_get = {{ function.hardcoded.get_function }}(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
        success_codes=[200],
    )

    if not response_get["result"]:
        if ctx.get("test", False):
            # Return a proposed state by Idem state --test
            result["new_state"] = hub.tool.azure.test_state_utils.generate_test_state(
                enforced_state={},
                # TODO Populate desired_state to match present input
                desired_state={"name": name,},
            )
            result["comment"]=("Would create azure.{{ function.hardcoded.create_ref }}",)
            return result

        if response_get["status"] == 404:
            # PUT operation to create a resource
            # TODO Populate payload, please refer to hub.tool.azure.virtual_networks.conversion_utils.convert_present_to_raw_virtual_network()
            payload = {}
            response_put = {{ function.hardcoded.create_function }}(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
                success_codes=[200, 201],
                json=payload,
            )

            if not response_put["result"]:
                hub.log.debug(f"Could not create {{ function.hardcoded.resource_name }} {response_put['comment']} {response_put['ret']}")
                result["comment"] = (response_put["comment"], response_put["ret"])
                result["result"] = False
                return result

            # TODO: Add conversion_util function here. Please refer to hub.tool.azure.virtual_networks.conversion_utils.convert_raw_virtual_network_to_present()
            result["new_state"] = {}
            result["comment"] = (f"Created azure.{{ function.hardcoded.create_ref }} '{name}'",)
            return result
        else:
            hub.log.debug(
                f"Could not get {{ function.hardcoded.resource_name }} {response_get['comment']} {response_get['ret']}"
            )
            result["result"] = False
            result["comment"] = (response_get["comment"], response_get["ret"]),
            return result
    else:
        existing_resource = response_get["ret"]
        # TODO: Add conversion_util function here. Please refer to hub.tool.azure.virtual_networks.conversion_utils.convert_raw_virtual_network_to_present
        result["old_state"] = {}
        # Generate a new PUT operation payload with new values
        # TODO: Add payload_update function here. Please refer to hub.exec.azure.virtual_networks.virtual_networks.update_virtual_network_payload()
        new_payload = {}
        if ctx.get("test", False):
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"] = (f"azure.{{ function.hardcoded.create_ref }} '{name}' has no property need to be updated.",)
            else:
                # TODO: Add conversion_util function here. Please refer to hub.tool.azure.virtual_networks.conversion_utils.convert_raw_virtual_network_to_present
                result["new_state"] = {}
                result["comment"] = (f"Would update azure.{{ function.hardcoded.create_ref }} '{name}'",)
            return result
        # PUT operation to update a resource
        if new_payload["ret"] is None:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"] = (
                f"azure.{{ function.hardcoded.create_ref }} '{name}' has no property need to be updated.",
            )
            return result
        result["comment"] = result["comment"] + new_payload["comment"]
        response_put = await hub.exec.request.json.put(
            ctx,
            url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
            success_codes=[200],
            json=new_payload["ret"],
        )

        if not response_put["result"]:
            hub.log.debug(
                f"Could not update azure.{{ function.hardcoded.create_ref }} '{name}' {response_put['comment']} {response_put['ret']}"
            )
            result["result"] = False
            result["comment"] = (response_get["comment"], response_get["ret"])
            return result

        # TODO: Add conversion_util function here. Please refer to hub.tool.azure.virtual_networks.conversion_utils.convert_raw_virtual_network_to_present
        result["new_state"] = {}
        result["comment"] = (f"Updated azure.{{ function.hardcoded.create_ref }} '{name}'",)
        return result
"""

ABSENT_REQUEST_FORMAT = r"""
    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    subscription_id = ctx.acct.subscription_id
    # TODO: This resource_id is not correct and it needs to be manually fixed to the correct one
    resource_id = {{ function.hardcoded.get_function }}
    # TODO: double check if the api version here is correct or not.
    response_get = {{ function.hardcoded.get_function }}(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
        success_codes=[200],
    )
    if response_get["result"]:
        # TODO: Add conversion_util function here. Please refer to hub.tool.azure.virtual_networks.conversion_utils.convert_raw_virtual_network_to_present
        result["old_state"] = {}
        if ctx.get("test", False):
            result["comment"] = ("Would delete azure.{{ function.hardcoded.create_ref }} '{name}'",)
            return result
        response_delete = {{ function.hardcoded.delete_function }}(
            ctx,
            url=f"{{ function.hardcoded.path }}",
            success_codes=[200, 202, 204],
        )

        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete azure.{{ function.hardcoded.create_ref }} '{name}' {response_delete['comment']} {response_delete['ret']}"
            )
            result["result"] = False
            result["comment"] = (response_delete["comment"], response_delete["ret"])
            return result

        result["comment"] = (
            f"Deleted azure.{{ function.hardcoded.create_ref }} '{name}'",
        )
        return result
    elif response_get["status"] == 404:
        # If Azure returns 'Not Found' error, it means the resource has been absent.
        result["comment"] = (f"azure.{{ function.hardcoded.create_ref }} '{name}' already absent",)
        return result
    else:
        hub.log.debug(
            f"Could not get azure.{{ function.hardcoded.create_ref }} '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = (response_get["comment"], response_get["ret"])
        return result
"""

DESCRIBE_REQUEST_FORMAT = r"""
    result = {}
    subscription_id = ctx.acct.subscription_id
    uri_parameters = {{ function.hardcoded.describe_parameters }}
    async for page_result in hub.tool.azure.request.paginate(
            ctx,
            url=f"{{ function.hardcoded.path }}",
            success_codes=[200],
    ):
        resource_list = page_result.get("value")
        if resource_list:
            for resource in resource_list:
                # TODO: double check if the resource_id is populated correctly
                resource_id = resource["id"]
                uri_parameter_values = (
                    hub.tool.azure.utils.get_uri_parameter_value_from_uri(resource_id, uri_parameters)
                )
                # TODO: Add conversion_util function here. Please refer to hub.tool.azure.virtual_networks.conversion_utils.convert_raw_virtual_network_to_present
                resource_translated = {}
                result[resource["id"]] = {f"azure.{{ function.hardcoded.create_ref }}.present":
                    [{parameter_key: parameter_value} for parameter_key, parameter_value in resource_translated.items()]}
    return result
"""



"""State module for managing Compute Virtual Machine."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
__reconcile_wait__ = {"static": {"wait_in_seconds": 10}}


async def present(
    hub,
    ctx,
    name: str,
    location: str,
    network_interface_ids: List[str],
    os_profile: make_dataclass(
        "OsProfile",
        [
            ("computer_name", str),
            ("admin_username", str),
            ("admin_password", str, field(default=None)),
            ("custom_data", str, field(default=None)),
            (
                "linux_configuration",
                make_dataclass(
                    "LinuxConfiguration",
                    [
                        (
                            "ssh_public_keys",
                            List[
                                make_dataclass(
                                    "SshPublicKey",
                                    [
                                        ("key_data", str, field(default=None)),
                                        ("path", str, field(default=None)),
                                    ],
                                )
                            ],
                            field(default=None),
                        ),
                        ("provision_vm_agent", bool, field(default=None)),
                        ("enable_vm_agent_platform_updates", bool, field(default=None)),
                        ("disable_password_authentication", bool, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "windows_configuration",
                make_dataclass(
                    "WindowsConfiguration",
                    [
                        ("provision_vm_agent", bool, field(default=None)),
                        ("enable_vm_agent_platform_updates", bool, field(default=None)),
                        ("enable_automatic_updates", bool, field(default=None)),
                        ("time_zone", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ),
    storage_os_disk: make_dataclass(
        "StorageOsDisk",
        [
            ("disk_name", str),
            ("disk_caching", str),
            ("disk_create_option", str),
            ("storage_account_type", str),
            ("disk_delete_option", str, field(default=None)),
            ("disk_size_in_GB", int, field(default=None)),
            ("disk_id", str, field(default=None)),
            ("disk_image_vhd_uri", str, field(default=None)),
            ("disk_os_type", str, field(default=None)),
            ("disk_vhd_uri", str, field(default=None)),
            ("write_accelerator_enabled", bool, field(default=None)),
        ],
    ),
    resource_id: str = None,
    subscription_id: str = None,
    resource_group_name: str = None,
    virtual_machine_name: str = None,
    virtual_machine_size: str = None,
    storage_image_reference: make_dataclass(
        "ImageReference",
        [
            ("image_sku", str, field(default=None)),
            ("image_publisher", str, field(default=None)),
            ("image_version", str, field(default=None)),
            ("image_offer", str, field(default=None)),
            ("image_id", str, field(default=None)),
            ("shared_gallery_image_id", str, field(default=None)),
            ("community_gallery_image_id", str, field(default=None)),
        ],
    ) = None,
    storage_data_disks: List[
        make_dataclass(
            "StorageDataDisks",
            [
                ("disk_create_option", str),
                ("disk_size_in_GB", int),
                ("disk_logical_unit_number", int),
                ("disk_name", str, field(default=None)),
                ("disk_caching", str, field(default=None)),
                ("disk_delete_option", str, field(default=None)),
                ("disk_id", str, field(default=None)),
                ("storage_account_type", str, field(default=None)),
            ],
        )
    ] = None,
    tags: Dict[str, str] = None,
    plan: make_dataclass(
        "Plan",
        [
            ("name", str),
            ("publisher", str),
            ("product", str),
            ("promotion_code", str, field(default=None)),
        ],
    ) = None,
    availability_set_id: str = None,
    license_type: str = None,
    boot_diagnostics: make_dataclass(
        "BootDiagnostics",
        [
            ("enabled", bool, field(default=None)),
            ("storage_uri", str, field(default=None)),
        ],
    ) = None,
    extensions_time_budget: str = None,
) -> Dict:
    r"""Create or update Virtual Machines.

    Args:
        name(str): The identifier for this state.
        location(str): Resource location. Changing this forces a new resource to be created.
        network_interface_ids(List[str]): A list of Network Interface IDs which should be associated with the Virtual Machine.
        os_profile(Dict[str, Any]): Specifies the operating system settings used while creating the virtual machine.

            * computer_name(str):
                Specifies the name of the Virtual Machine.
            * admin_username(str):
                Specifies the name of the local administrator account.
            * admin_password(str, Optional):
                (Required for Windows, Optional for Linux) The password associated with the local administrator account.
            * custom_data(str, Optional):
                Specifies a base-64 encoded string of custom data. The base-64 encoded string is decoded to a binary array that is saved as a file on the Virtual Machine. The maximum length of the binary array is 65535 bytes. Note: Do not pass any secrets or passwords in customData property. This property cannot be updated after the VM is created.
            * linux_configuration(Dict[str, Any], Optional):
                Specifies the Linux operating system settings on the virtual machine.

                * ssh_public_keys(List[Dict[str, Any]], Optional):
                    The list of SSH public keys used to authenticate with linux based VMs.

                    * key_data(str, Optional):
                        SSH public key certificate used to authenticate with the VM through ssh. The key needs to be at least 2048-bit and in ssh-rsa format.
                    * path(str, Optional):
                        Specifies the full path on the created VM where ssh public key is stored. If the file already exists, the specified key is appended to the file. Example: /home/user/.ssh/authorized_keys
                * provision_vm_agent(bool, Optional):
                    Indicates whether virtual machine agent should be provisioned on the virtual machine. When this property is not specified in the request body, default behavior is to set it to true. This will ensure that VM Agent is installed on the VM so that extensions can be added to the VM later.
                * enable_vm_agent_platform_updates(bool, Optional):
                    Indicates whether VMAgent Platform Updates is enabled for the Linux virtual machine. Default value is false.
                * disable_password_authentication(bool, Optional):
                    Specifies whether password authentication should be disabled.
            * windows_configuration(Dict[str, Any], Optional):
                Specifies Windows operating system settings on the virtual machine.

                * provision_vm_agent(bool, Optional):
                    Indicates whether virtual machine agent should be provisioned on the virtual machine. When this property is not specified in the request body, it is set to true by default. This will ensure that VM Agent is installed on the VM so that extensions can be added to the VM later.
                * enable_vm_agent_platform_updates(bool, Optional):
                    Indicates whether VMAgent Platform Updates is enabled for the Windows virtual machine. Default value is false.
                * enable_automatic_updates(bool, Optional):
                    Indicates whether Automatic Updates is enabled for the Windows virtual machine. Default value is true.
                * time_zone(str, Optional):
                    Specifies the time zone of the virtual machine. e.g. "Pacific Standard Time". Possible values can be TimeZoneInfo.Id value from time zones returned by TimeZoneInfo.GetSystemTimeZones.
        storage_os_disk(Dict[str, Any]): Specifies information about the operating system disk used by the virtual machine.

            * disk_name(str):
                Specifies the name of the OS Disk.
            * disk_caching(str, Optional):
                Specifies the caching requirements for the OS Disk. Possible values include None, ReadOnly and ReadWrite.
            * disk_create_option(str):
                Specifies how the OS Disk should be created. Possible values are Attach (managed disks only) and FromImage.
            * storage_account_type(str):
                Specifies the type of Managed Disk which should be created. Possible values are Standard_LRS, StandardSSD_LRS or Premium_LRS.
            * disk_delete_option(str, Optional):
                Specifies how the OS Disk should be handled after VM deletion. Possible values are Detach and Delete.
            * disk_size_in_GB(str, Optional):
                Specifies the size of the OS Disk in gigabytes.
            * disk_id(str, Optional):
                Specifies the ID of an existing Managed Disk which should be attached as the OS Disk of this Virtual Machine. If this is set then the create_option must be set to Attach.
            * disk_image_vhd_uri(str, Optional):
                The source user image virtual hard disk. The virtual hard disk will be copied before being attached to the virtual machine. If SourceImage is provided, the destination virtual hard drive must not exist. Specifies the virtual hard disk's uri.
            * disk_os_type(str, Optional):
                This property allows you to specify the type of the OS that is included in the disk if creating a VM from user-image or a specialized VHD. Possible values are: Windows, Linux.
            * disk_vhd_uri(str, Optional):
                Specifies the virtual hard disk's uri.
            * write_accelerator_enabled(bool, Optional):
                Specifies whether writeAccelerator should be enabled or disabled on the disk.
        resource_id(str, Optional): Virtual Machine resource id on Azure
        subscription_id(str, Optional): Subscription Unique id.
        resource_group_name(str, Optional): The name of the resource group.
        virtual_machine_name(str, Optional): The name of the virtual machine.
        virtual_machine_size(str, Optional): Specifies the size of the Virtual Machine.
        storage_image_reference(Dict[str, Any], Optional): Specifies information about the image to use. Eg- platform images, marketplace images.

            * image_sku(str, Optional):
                Specifies the SKU of the image used to create the virtual machine. Changing this forces a new resource to be created.
            * image_publisher(str, Optional):
                Specifies the publisher of the image used to create the virtual machine. Changing this forces a new resource to be created.
            * image_version(str, Optional):
                Specifies the version of the image used to create the virtual machine. Changing this forces a new resource to be created.
            * image_offer(str, Optional):
                Specifies the offer of the image used to create the virtual machine. Changing this forces a new resource to be created.
            * image_id(str, Optional):
                Resource Id
            * shared_gallery_image_id(str, Optional):
                Specified the shared gallery image unique id for vm deployment. This can be fetched from shared gallery image GET call.
            * community_gallery_image_id(str, Optional):
                Specified the community gallery image unique id for vm deployment. This can be fetched from community gallery image GET call.
        storage_data_disks(List[Dict[str, Any]], Optional): List of Data disks attached/added to a VM.

            * disk_create_option(str):
                Specifies how the data disk should be created. Possible values are Attach, FromImage and Empty.
            * disk_size_in_GB(int):
                Specifies the size of the data disk in gigabytes.
            * disk_logical_unit_number(int):
                Specifies the logical unit number of the data disk. This needs to be unique within all the Data Disks on the Virtual Machine.
            * disk_name(str, Optional):
                The name of the Data Disk.
            * disk_caching(str, Optional):
                Specifies the caching requirements for the Data Disk. Possible values include None, ReadOnly and ReadWrite.
            * disk_delete_option(str, Optional):
                Specifies how the OS Disk should be handled after VM deletion. Possible values are Detach and Delete.
            * disk_id(str, Optional):
                Specifies the ID of an Existing Managed Disk which should be attached to this Virtual Machine. When this field is set create_option must be set to Attach.
            * storage_account_type(str, Optional):
                Specifies the type of managed disk to create. Possible values are either Standard_LRS, StandardSSD_LRS, Premium_LRS or UltraSSD_LRS.
        tags(Dict[str, str], Optional): Resource tags.
        plan(Dict[str, str], Optional): Specifies information about the marketplace image used to create the virtual machine. This element is only used for marketplace images. Before you can use a marketplace image from an API, you must enable the image for programmatic use.  In the Azure portal, find the marketplace image that you want to use and then click **Want to deploy programmatically, Get Started ->**. Enter any required information and then click **Save**.

            * name(str):
                The plan ID.
            * publisher(str):
                The publisher ID.
            * product(str):
                Specifies the product of the image from the marketplace. This is the same value as Offer under the imageReference element.
            * promotion_code(str, Optional):
                The promotion code.
        availability_set_id(str, Optional): Specifies id of the availability set that the virtual machine should be assigned to. Virtual machines specified in the same availability set are allocated to different nodes to maximize availability. Currently, a VM can only be added to availability set at creation time. The availability set to which the VM is being added should be under the same resource group as the availability set resource. An existing VM cannot be added to an availability set. This property cannot exist along with a non-null properties.virtualMachineScaleSet reference.
        license_type(str, Optional): Specifies that the image or disk that is being used was licensed on-premises. Possible values for Windows Server operating system are: Windows_Client, Windows_Server. Possible values for Linux Server operating system are: RHEL_BYOS (for RHEL), SLES_BYOS (for SUSE). For more information, see Azure Hybrid Use Benefit for Windows Server and Azure Hybrid Use Benefit for Linux Server
        boot_diagnostics(Dict[str, Any], Optional): Specifies the boot diagnostic settings state. Boot Diagnostics is a debugging feature which allows you to view Console Output and Screenshot to diagnose VM status.

            * enabled(bool, Optional):
                Whether boot diagnostics should be enabled on the Virtual Machine.
            * storage_uri(str, Optional):
                Uri of the storage account to use for placing the console output and screenshot. If storageUri is not specified while enabling boot diagnostics, managed storage will be used. NOTE: If storageUri is being specified then ensure that the storage account is in the same region and subscription as the VM.
        extensions_time_budget(str, Optional): Specifies the time alloted for all extensions to start. The time duration should be between 15 minutes and 120 minutes (inclusive) and should be specified in ISO 8601 format. The default value is 90 minutes (PT1H30M).

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              azure.compute.virtual_machines.present:
                - name: my-vm
                - resource_group_name: my-rg-1
                - virtual_machine_name: my-vm
                - location: eastus
                - virtual_machine_size: Standard_B1ls
                - network_interface_ids:
                  - /subscriptions/subscription_id/resourceGroups/my-rg-1/providers/Microsoft.Network/networkInterfaces/my-nic-id-1
                - storage_image_reference:
                    image_sku: 18.04-LTS
                    image_publisher: Canonical
                    image_version: latest
                    image_offer: UbuntuServer
                - storage_os_disk:
                    storage_account_type: Standard_LRS
                    disk_name: my-os-disk
                    disk_caching: ReadWrite
                    disk_size_in_GB: 30
                    disk_create_option: FromImage
                    disk_delete_option: Detach
                - storage_data_disks:
                  - disk_name: my-data-disk
                    disk_size_in_GB: 2
                    disk_logical_unit_number: 0
                    disk_caching: None
                    disk_create_option: Empty
                    disk_delete_option: Delete
                - os_profile:
                    admin_username: my-admin-username
                    computer_name: machine-name
                    admin_password: Vmwareadmin123!
                - tags:
                    my-tag-key-1: my-tag-value-1
                    my-tag-key-2: my-tag-value-2
    """
    result = ctx.get("wrapper_result")

    if ctx.get("skip_present"):
        return result

    if not result:
        error_message = hub.tool.azure.comment_utils.no_result_from_wrapper(
            "azure.compute.virtual_machines", name
        )
        hub.log.error(error_message)
        return {
            "result": False,
            "old_state": None,
            "new_state": None,
            "name": name,
            "comment": [error_message],
        }

    computed = ctx.get("computed")
    computed_resource_id = computed.get("resource_id")
    computed_resource_url = computed.get("resource_url")

    existing_resource = result["old_state"]
    if not existing_resource:
        if ctx.get("test", False):
            # Return a proposed state by Idem state --test
            result["new_state"] = hub.tool.azure.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_group_name": resource_group_name,
                    "subscription_id": subscription_id,
                    "virtual_machine_name": virtual_machine_name,
                    "virtual_machine_size": virtual_machine_size,
                    "tags": tags,
                    "location": location,
                    "network_interface_ids": network_interface_ids,
                    "storage_image_reference": storage_image_reference,
                    "storage_os_disk": storage_os_disk,
                    "storage_data_disks": storage_data_disks,
                    "os_profile": os_profile,
                    "resource_id": computed_resource_id,
                    "plan": plan,
                    "availability_set_id": availability_set_id,
                    "license_type": license_type,
                    "boot_diagnostics": boot_diagnostics,
                    "extensions_time_budget": extensions_time_budget,
                    "status": "running",
                },
            )
            result["comment"].append(
                f"Would create azure.compute.virtual_machines '{name}'"
            )
            return result

        else:
            # PUT operation to create a resource
            payload = hub.tool.azure.compute.virtual_machines.convert_present_to_raw_virtual_machine(
                location=location,
                virtual_machine_size=virtual_machine_size,
                network_interface_ids=network_interface_ids,
                storage_image_reference=storage_image_reference,
                storage_os_disk=storage_os_disk,
                storage_data_disks=storage_data_disks,
                os_profile=os_profile,
                tags=tags,
                plan=plan,
                availability_set_id=availability_set_id,
                license_type=license_type,
                boot_diagnostics=boot_diagnostics,
                extensions_time_budget=extensions_time_budget,
            )

            response_put = await hub.exec.request.json.put(
                ctx,
                url=computed_resource_url,
                success_codes=[200, 201],
                json=payload,
            )
            if not response_put["result"]:
                hub.log.debug(
                    f"Could not create azure.compute.virtual_machines {response_put['comment']} {response_put['ret']}"
                )
                result["comment"].extend(
                    hub.tool.azure.result_utils.extract_error_comments(response_put)
                )
                result["result"] = False
                return result

            result[
                "new_state"
            ] = hub.tool.azure.compute.virtual_machines.convert_raw_virtual_machine_to_present(
                resource=response_put["ret"],
                idem_resource_name=name,
                subscription_id=subscription_id,
                resource_group_name=resource_group_name,
                virtual_machine_name=virtual_machine_name,
                resource_id=computed_resource_id,
            )
            result["comment"].append(f"Created azure.compute.virtual_machines '{name}'")

            # Setting it to the initial state the VM is in after create has completed successfully
            result["new_state"]["status"] = "starting"

            return result
    else:
        raw_existing_resource = hub.tool.azure.compute.virtual_machines.convert_present_to_raw_virtual_machine(
            **existing_resource
        )
        # Generate a new PUT operation payload with new values
        new_payload = (
            hub.tool.azure.compute.virtual_machines.update_virtual_machine_payload(
                raw_existing_resource,
                existing_resource,
                {
                    "virtual_machine_size": virtual_machine_size,
                    "storage_data_disks": storage_data_disks,
                    "network_interface_ids": network_interface_ids,
                    "storage_os_disk": storage_os_disk,
                    "os_profile": os_profile,
                    "storage_image_reference": storage_image_reference,
                    "tags": tags,
                    "plan": plan,
                    "availability_set_id": availability_set_id,
                    "license_type": license_type,
                    "boot_diagnostics": boot_diagnostics,
                    "extensions_time_budget": extensions_time_budget,
                },
            )
        )

        if new_payload["ret"] is None:
            result["new_state"] = copy.deepcopy(result["old_state"])
            result["comment"].append(
                hub.tool.azure.comment_utils.no_property_to_be_updated_comment(
                    "azure.compute.virtual_machines", name
                )
            )
            return result

        if ctx.get("test", False):
            result[
                "new_state"
            ] = hub.tool.azure.compute.virtual_machines.convert_raw_virtual_machine_to_present(
                resource=new_payload["ret"],
                idem_resource_name=name,
                subscription_id=subscription_id,
                resource_group_name=resource_group_name,
                virtual_machine_name=virtual_machine_name,
                resource_id=computed_resource_id,
            )
            # The VM status is read only and gets returned only when get is called
            # Its value will not be changed when present is called,
            # so we copy the old state to the new state at this point
            # It is only used to manage the state with the exec functions
            # It can be used from consumers to perform VM operations based on its value
            if result["old_state"].get("status"):
                result["new_state"]["status"] = result["old_state"]["status"]
            result["comment"].append(
                f"Would update azure.compute.virtual_machines '{name}'"
            )
            return result

        # PUT operation to update a resource
        response_put = await hub.exec.request.json.put(
            ctx,
            url=computed_resource_url,
            success_codes=[200, 201],
            json=new_payload["ret"],
        )
        if not response_put["result"]:
            hub.log.debug(
                f"Could not update azure.compute.virtual_machines {response_put['comment']} {response_put['ret']}"
            )
            result["result"] = False
            result["comment"].extend(
                hub.tool.azure.result_utils.extract_error_comments(response_put)
            )
            return result

        result[
            "new_state"
        ] = hub.tool.azure.compute.virtual_machines.convert_raw_virtual_machine_to_present(
            resource=response_put["ret"],
            idem_resource_name=name,
            subscription_id=subscription_id,
            resource_group_name=resource_group_name,
            virtual_machine_name=virtual_machine_name,
            resource_id=computed_resource_id,
        )

        # The VM status is read only and gets returned only when get is called
        # Its value will not be changed when present is called,
        # so we copy the old state to the new state at this point
        # It is only used to manage the state with the exec functions
        # It can be used from consumers to perform VM operations based on its value
        if result["old_state"].get("status"):
            result["new_state"]["status"] = result["old_state"]["status"]

        result["comment"].append(f"Updated azure.compute.virtual_machines '{name}'")
        return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_group_name: str = None,
    virtual_machine_name: str = None,
    subscription_id: str = None,
    resource_id: str = None,
) -> Dict:
    r"""Delete a Virtual Machine.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str, Optional): The name of the resource group.
        virtual_machine_name(str, Optional): The name of the virtual machine.
        subscription_id(str, Optional): Subscription Unique id.
        resource_id(str, Optional): An identifier of the resource in the provider. Defaults to None.

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.compute.virtual_machines.absent:
                - name: my-vm
                - resource_group_name: my-resource-group
                - virtual_machine_name: my-vm
                - subscription_id: my-subscription
    """

    # Resource deletion is handled via common recursive_contracts.init#call_absent() wrapper
    raise NotImplementedError


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all Virtual Machines under the same subscription.


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.compute.virtual_machines
    """
    result = {}
    ret_list = await hub.exec.azure.compute.virtual_machines.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(
            f"Could not describe compute virtual_machines {ret_list['comment']}"
        )
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.compute.virtual_machines.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """Default implemented for each module."""
    return hub.tool.azure.resource_utils.is_pending(
        ret=ret, state=state, **pending_kwargs
    )



import copy
from typing import Any
from typing import Dict
from typing import List


def update_subnets_payload(
    hub, existing_payload: Dict[str, Any], new_values: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
     PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
         this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
         do an update.

    Returns:
        A result dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages tuple.
    """
    result = {"result": True, "ret": None, "comment": []}
    need_update = False
    new_payload = copy.deepcopy(existing_payload)
    existing_properties = existing_payload["properties"]
    if (new_values.get("address_prefix") is not None) and (
        new_values["address_prefix"] != existing_properties.get("addressPrefix")
    ):
        new_payload["properties"]["addressPrefix"] = new_values.get("address_prefix")
        need_update = True
    if (
        new_values.get("enforce_private_link_endpoint_network_policies") is not None
    ) and (
        compare_policies(
            new_values["enforce_private_link_endpoint_network_policies"],
            existing_properties.get("privateEndpointNetworkPolicies"),
        )
    ):
        if new_values.get("enforce_private_link_endpoint_network_policies") is True:
            new_payload["properties"]["privateEndpointNetworkPolicies"] = "Enabled"
        elif new_values.get("enforce_private_link_endpoint_network_policies") is False:
            new_payload["properties"]["privateEndpointNetworkPolicies"] = "Disabled"
        need_update = True
    if new_values.get("delegations") is not None:
        if compare_delegations(
            new_values.get("delegations"),
            convert_raw_delegations_to_present(existing_properties.get("delegations")),
            "name",
        ):
            new_payload["properties"][
                "delegations"
            ] = convert_present_to_raw_delegations(new_values["delegations"])
            need_update = True
    if (
        new_values.get("enforce_private_link_service_network_policies") is not None
    ) and (
        compare_policies(
            new_values["enforce_private_link_service_network_policies"],
            existing_properties.get("privateLinkServiceNetworkPolicies"),
        )
    ):
        if new_values.get("enforce_private_link_service_network_policies") is True:
            new_payload["properties"]["privateLinkServiceNetworkPolicies"] = "Enabled"
        elif new_values.get("enforce_private_link_service_network_policies") is False:
            new_payload["properties"]["privateLinkServiceNetworkPolicies"] = "Disabled"
        need_update = True
    if (new_values.get("service_endpoints") is not None) and (
        (existing_properties.get("serviceEndpoints") is None)
        or (
            set(new_values["service_endpoints"])
            != {
                service_endpoint["service"]
                for service_endpoint in existing_properties.get("serviceEndpoints")
            }
        )
    ):
        new_payload["properties"]["serviceEndpoints"] = [
            {"service": service_name}
            for service_name in new_values["service_endpoints"]
        ]
        need_update = True
    if (new_values.get("service_endpoint_policy_ids") is not None) and (
        (existing_properties.get("serviceEndpointPolicies") is None)
        or (
            set(new_values["service_endpoint_policy_ids"])
            != {
                policy["id"]
                for policy in existing_properties.get("serviceEndpointPolicies")
            }
        )
    ):
        new_payload["properties"]["serviceEndpointPolicies"] = [
            {"id": policy_id} for policy_id in new_values["service_endpoint_policy_ids"]
        ]
        need_update = True
    if need_update:
        result["ret"] = new_payload
    return result


def convert_raw_subnets_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_group_name: str,
    virtual_network_name: str,
    subnet_name: str,
    resource_id: str,
    subscription_id: str = None,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        virtual_network_name: Azure Virtual Network resource name.
        subnet_name: Azure Subnet Resource name.
        resource_id: Azure Subnet resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "virtual_network_name": virtual_network_name,
        "subnet_name": subnet_name,
        "subscription_id": subscription_id,
    }
    properties = resource.get("properties")
    if properties:
        if properties.get("addressPrefix") is not None:
            resource_translated["address_prefix"] = properties.get("addressPrefix")
        if properties.get("privateEndpointNetworkPolicies") is not None:
            if properties.get("privateEndpointNetworkPolicies") == "Enabled":
                resource_translated[
                    "enforce_private_link_endpoint_network_policies"
                ] = True
            elif properties.get("privateEndpointNetworkPolicies") == "Disabled":
                resource_translated[
                    "enforce_private_link_endpoint_network_policies"
                ] = False
        if properties.get("privateLinkServiceNetworkPolicies") is not None:
            if properties.get("privateLinkServiceNetworkPolicies") == "Enabled":
                resource_translated[
                    "enforce_private_link_service_network_policies"
                ] = True
            elif properties.get("privateLinkServiceNetworkPolicies") == "Disabled":
                resource_translated[
                    "enforce_private_link_service_network_policies"
                ] = False
        if len(properties.get("delegations")) != 0:
            resource_translated["delegations"] = convert_raw_delegations_to_present(
                properties.get("delegations")
            )
        if properties.get("serviceEndpoints") is not None:
            resource_translated["service_endpoints"] = [
                service_endpoints["service"]
                for service_endpoints in properties.get("serviceEndpoints")
            ]
        if properties.get("serviceEndpointPolicies") is not None:
            resource_translated["service_endpoint_policy_ids"] = [
                service_endpoint_policy["id"]
                for service_endpoint_policy in properties.get("serviceEndpointPolicies")
            ]

    return resource_translated


def convert_present_to_raw_subnets(
    hub,
    address_prefix: str,
    enforce_private_link_endpoint_network_policies: bool = None,
    delegations: List = None,
    enforce_private_link_service_network_policies: bool = None,
    service_endpoints: List = None,
    service_endpoint_policy_ids: List = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        address_prefix: The address prefix for the subnet.
        enforce_private_link_endpoint_network_policies: Enable or Disable apply network policies on private end point in the subnet.
        delegations: An array of references to the delegations on the subnet.
        enforce_private_link_service_network_policies: Enable or Disable apply network policies on private link service in the subnet.
        service_endpoints: An array of service endpoints
        service_endpoint_policy_ids: An array of service endpoint policy Ids.
    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {
        "properties": {"addressPrefix": address_prefix},
    }
    if enforce_private_link_endpoint_network_policies is not None:
        if enforce_private_link_endpoint_network_policies is True:
            payload["properties"]["privateEndpointNetworkPolicies"] = "Enabled"
        elif enforce_private_link_endpoint_network_policies is False:
            payload["properties"]["privateEndpointNetworkPolicies"] = "Disabled"
    if delegations is not None:
        payload["properties"]["delegations"] = convert_present_to_raw_delegations(
            delegations
        )
    if enforce_private_link_service_network_policies is not None:
        if enforce_private_link_service_network_policies is True:
            payload["properties"]["privateLinkServiceNetworkPolicies"] = "Enabled"
        elif enforce_private_link_service_network_policies is False:
            payload["properties"]["privateLinkServiceNetworkPolicies"] = "Disabled"
    if service_endpoints is not None:
        payload["properties"]["serviceEndpoints"] = [
            {"service": service_name} for service_name in service_endpoints
        ]
    if service_endpoint_policy_ids is not None:
        payload["properties"]["serviceEndpointPolicies"] = [
            {"id": policy_id} for policy_id in service_endpoint_policy_ids
        ]

    return payload


def convert_raw_delegations_to_present(delegations: List):
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
    present input parameters.

    Args:
        delegations(list[dict], optional): Resource List of Delegation in a subnet resource.

    Returns:
         A delegation List that contains the parameters that match respective present function's input format.
    """
    present_delegations: List = []
    for delegation in delegations:
        new_delegation_payload = {
            "name": delegation["name"],
            "service": delegation["properties"]["serviceName"],
        }
        present_delegations.append(new_delegation_payload)
    return present_delegations


def convert_present_to_raw_delegations(delegations: List):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        delegations(list[dict], required): List of Delegation in a subnet resource.Each Delegation will have fields
        ((name(str), required) and service(str), required))

    Returns:
        Delegations List(Dict[str,any]) in the format of an Azure PUT operation payload.
    """
    raw_delegations: List = []
    for delegation in delegations:
        new_delegation_payload = {
            "name": delegation["name"],
            "properties": {"serviceName": delegation["service"]},
        }
        raw_delegations.append(new_delegation_payload)
    return raw_delegations


def compare_delegations(
    new_values: List[Dict[str, any]], resource: List[Dict[str, any]], sorting_key: str
):
    """
    Compares subnets delegations to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        new_values: (List[Dict[str, any]]) Present value which will be given as input
        resource: (List[Dict[str, any]]) Raw resource response which needs to be compared with new_values
        sorting_key: (str) Primary/Unique key name within each dictionary , which will be used to sort dictionary
         objects with given list before comparing.

    Returns:
        A boolean value, True if there is any difference between List[Dict] arguments else returns False
    """
    return sorted(new_values, key=lambda x: x[sorting_key]) != sorted(
        resource, key=lambda x: x[sorting_key]
    )


def compare_policies(new_values: bool, resource: str):
    """
    Compares subnets service policies and subnets endpoint policies to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        new_values: (bool) Present value which will be given as input
        resource: (str) Raw resource response which needs to be compared with new_values

    Returns:
        A boolean value, True if there is any difference between arguments else returns False
    """
    if (resource == "Disabled" and new_values == True) or (
        resource == "Enabled" and new_values == False
    ):
        return True
    return False



"""Exec module for managing Network Firewall"""

__func_alias__ = {"list_": "list"}

from typing import Dict
from collections import OrderedDict


async def get(hub, ctx, resource_id: str, name: str = None, raw: bool = False) -> Dict:
    """Gets network Firewall from azure account.

    Args:
        resource_id(str):
            The resource id of the Firewall.
        name(str, Optional):
            The name of the resource
        raw(bool, Optional):
            Returns raw response if True. Defaults to False

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec azure.network.firewall.get resource_id="value" raw="False"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path:  azure.network.firewall.get
                - kwargs:
                    resource_id: "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/azureFirewalls/{firewall_name}"
                    raw: False
    """

    result = dict(comment=[], ret=None, result=True)
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2022-07-01",
        success_codes=[200],
    )
    if not response_get["result"]:
        if response_get["status"] != 404:
            result["result"] = False
        result["comment"].extend(
            hub.tool.azure.result_utils.extract_error_comments(response_get)
        )
        return result
    uri_parameters = OrderedDict(
        {
            "subscriptions": "subscription_id",
            "resourceGroups": "resource_group_name",
            "azureFirewalls": "firewall_name",
        }
    )
    uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
        resource_id, uri_parameters
    )
    if raw:
        result["ret"] = response_get["ret"]
    else:
        result["ret"] = hub.tool.azure.network.firewall.convert_raw_firewall_to_present(
            idem_resource_name=resource_id,
            resource=response_get["ret"],
            resource_id=resource_id,
            **uri_parameter_values,
        )
    return result


async def list_(
    hub,
    ctx,
) -> Dict:
    """Lists all Network Firewall.

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec azure.network.firewall.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: azure.network.firewall.list

    """

    result = dict(comment=[], ret=[], result=True)
    subscription_id = ctx.acct.subscription_id
    uri_parameters = OrderedDict(
        {
            "subscriptions": "subscription_id",
            "resourceGroups": "resource_group_name",
            "azureFirewalls": "firewall_name",
        }
    )
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/providers/Microsoft.Network/azureFirewalls?api-version=2022-07-01",
        success_codes=[200],
    ):
        resource_list = page_result.get("value", None)
        if resource_list:
            for resource in resource_list:
                resource_id = resource["id"]
                uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
                    resource_id, uri_parameters
                )
                result["ret"].append(
                    hub.tool.azure.network.firewall.convert_raw_firewall_to_present(
                        resource=resource,
                        idem_resource_name=resource_id,
                        resource_id=resource_id,
                        **uri_parameter_values,
                    )
                )
    return result
